from __future__ import annotations

import logging

from prettyqt import constants, core


logger = logging.getLogger(__name__)


class MeltProxyModel(core.AbstractProxyModel):
    """Proxy model to unpivot a table from wide format to long format.

    Works same way as [pandas.melt](https://shorturl.at/bhGI3).

    === "Without proxy"

        ```py
        app = widgets.app()
        data = dict(
            first=["John", "Mary"],
            last=["Doe", "Bo"],
            height=[5.5, 6.0],
            weight=[130, 150],
        )
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        # table.proxifier.melt(id_columns=[0, 1])
        table.show()

        ```
        <figure markdown>
          ![Image title](../../images/meltproxymodel_before.png)
        </figure>

    === "With proxy"

        ```py
        app = widgets.app()
        data = dict(
            first=["John", "Mary"],
            last=["Doe", "Bo"],
            height=[5.5, 6.0],
            weight=[130, 150],
        )
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.proxifier.melt(id_columns=[0, 1])
        table.show()
        ```
        <figure markdown>
          ![Image title](../../images/meltproxymodel_after.png)
        </figure>

    ```py
    table.proxifier.melt(id_columns=[0, 1])
    # equals
    proxy = itemmodels.MeltProxyModel(id_columns=[0, 1])
    proxy.set_source_model(table.model())
    table.set_model(proxy)
    ```
    """

    ID = "melt"
    ICON = "mdi6.table-pivot"

    def __init__(
        self,
        id_columns: list[int],
        var_name: str = "Variable",
        value_name: str = "Value",
        **kwargs,
    ):
        self._id_columns = id_columns
        self._var_name = var_name
        self._value_name = value_name
        super().__init__(**kwargs)

    @property
    def value_columns(self) -> list[int]:
        colcount = self.sourceModel().columnCount()
        return [i for i in range(colcount) if i not in self._id_columns]

    def rowCount(self, index: core.ModelIndex | None = None) -> int:
        return self.sourceModel().rowCount() * len(self.value_columns)

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        parent = parent or core.ModelIndex()
        return 0 if self.sourceModel() is None else len(self._id_columns) + 2

    def is_source_column(self, column: int) -> bool:
        return 0 <= column < self.columnCount() - 2

    def is_variable_column(self, column: int) -> bool:
        return column == self.columnCount() - 2

    def is_value_column(self, column: int) -> bool:
        return column == self.columnCount() - 1

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        column = index.column()
        if self.is_variable_column(column) and role == constants.DISPLAY_ROLE:
            col = index.row() // self.sourceModel().rowCount()
            return self.sourceModel().headerData(
                self.value_columns[col], constants.HORIZONTAL
            )
        return super().data(index, role)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if orientation != constants.HORIZONTAL:
            return str(section)
        if self.is_variable_column(section):
            return self._var_name or "Variable"
        elif self.is_value_column(section):
            return self._value_name or "Value"
        else:
            section = self.get_source_column_for_proxy_column(section)
            return self.sourceModel().headerData(section, orientation, role)

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        # TODO: broken
        parent = parent or core.ModelIndex()
        if column not in self._id_columns:
            return self.createIndex(row, column, core.ModelIndex())
        col_pos = self.get_source_column_for_proxy_column(column)
        row_pos = row % self.sourceModel().rowCount()
        return self.sourceModel().index(row_pos, col_pos, parent)

    def parent(self, index: core.ModelIndex):
        if not self.is_source_column(index.column()):
            return core.ModelIndex()
        return self.sourceModel().parent(index)

    def get_source_column_for_proxy_column(self, column: int) -> int:
        return self._id_columns.index(column)

    def get_proxy_column_for_source_column(self, column: int) -> int:
        return column - sum(column > col for col in self._id_columns)

    def mapToSource(self, proxy_index: core.ModelIndex) -> core.ModelIndex:
        source = self.sourceModel()
        if source is None or not proxy_index.isValid():
            return core.ModelIndex()
        row, column = proxy_index.row(), proxy_index.column()
        row_count = source.rowCount()
        if self.is_variable_column(column):
            return core.ModelIndex()
        elif self.is_value_column(column):
            source_col = self.value_columns[row // row_count]
            source_row = row % row_count
            return source.index(source_row, source_col, core.ModelIndex())
        else:
            source_col = self.get_source_column_for_proxy_column(column)
            source_row = row % row_count
            return source.index(source_row, source_col)

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        # TODO: this is still broken.
        source = self.sourceModel()
        if source is None or not source_index.isValid():
            return core.ModelIndex()
        row, col = source_index.row(), source_index.column()
        # we can only really return a corresponding index for the value columns.
        # Var column is completely virtual and the id columns would have multiple
        # source indexes which correspond to the proxy index.
        if col not in self.value_columns:
            return core.ModelIndex()
        # TODO: convert row / col
        return source.index(row, col, core.ModelIndex())

    def get_id_columns(self) -> list[int]:
        """Get list of identifier columns."""
        return self._id_columns

    def set_id_columns(self, columns: list[int]):
        """Set identifier variable columns."""
        with self.reset_model():
            self._id_columns = columns

    def get_var_name(self) -> str:
        """Get variable column header."""
        return self._var_name

    def set_var_name(self, name: str):
        """Set header for variable column."""
        self._var_name = name
        section = self.columnCount() - 2
        self.headerDataChanged.emit(constants.HORIZONTAL, section, section)

    def get_value_name(self) -> str:
        """Get value column header."""
        return self._value_name

    def set_value_name(self, name: str):
        """Set header for value column."""
        self._value_name = name
        section = self.columnCount() - 1
        self.headerDataChanged.emit(constants.HORIZONTAL, section, section)

    id_columns = core.Property(
        list,
        get_id_columns,
        set_id_columns,
        doc="Columns to use as identifier variables",
    )
    var_name = core.Property(
        str,
        get_var_name,
        set_var_name,
        doc="Header for variable column",
    )
    value_name = core.Property(
        str,
        get_value_name,
        set_value_name,
        doc="Header for value column",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app(style="Windows")
    data = dict(
        first=["John", "Mary"],
        last=["Doe", "Bo"],
        height=[5.5, 6.0],
        weight=[130, 150],
    )
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.setWindowTitle("MeltProxyModel example")
    table.set_icon("mdi6.table-pivot")
    table.show()
    table.resize(600, 130)
    table.h_header.resize_sections("stretch")
    table.proxifier.melt(id_columns=[0, 1])
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import os

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict, datatypes


MatchModeStr = Literal["default", "extension", "content"]

MATCH_MODE: bidict[MatchModeStr, core.QMimeDatabase.MatchMode] = bidict(
    default=core.QMimeDatabase.MatchMode.MatchDefault,
    extension=core.QMimeDatabase.MatchMode.MatchExtension,
    content=core.QMimeDatabase.MatchMode.MatchContent,
)


class MimeDatabase(core.QMimeDatabase):
    """Maintains a database of MIME types."""

    def get_mime_type_for_file(
        self,
        path: datatypes.PathType | core.QFileInfo,
        match_mode: MatchModeStr | core.QMimeDatabase.MatchMode = "default",
    ) -> core.MimeType:
        if isinstance(path, os.PathLike):
            path = os.fspath(path)
        mime_type = self.mimeTypeForFile(path, MATCH_MODE.get_enum_value(match_mode))
        return core.MimeType(mime_type)

    def get_mime_type_for_data(
        self, data: datatypes.ByteArrayType | core.QIODevice
    ) -> core.MimeType:
        return core.MimeType(self.mimeTypeForData(data))

    def get_mime_type_for_filename_and_data(
        self, filename: os.PathLike, data: datatypes.ByteArrayType | core.QIODevice
    ) -> core.MimeType:
        path = os.fspath(filename)
        return core.MimeType(self.mimeTypeForFileNameAndData(path, data))

    def get_mime_type_for_name(self, name: str) -> core.MimeType:
        return core.MimeType(self.mimeTypeForName(name))

    def get_mime_type_for_url(self, url: core.QUrl | str) -> core.MimeType:
        url = core.QUrl(url) if isinstance(url, str) else url
        return core.MimeType(self.mimeTypeForUrl(url))

    def get_mime_types_for_filename(
        self, filename: datatypes.PathType
    ) -> list[core.MimeType]:
        path = os.fspath(filename)
        return [core.MimeType(i) for i in self.mimeTypesForFileName(path)]


if __name__ == "__main__":
    db = MimeDatabase()
    print(db.get_mime_type_for_file("C:/test.log"))



from __future__ import annotations

from collections.abc import Callable
from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict, get_repr, serializemixin


TypeStr = Literal[
    "linear",
    "in_quad",
    "out_quad",
    "in_out_quad",
    "out_in_quad",
    "in_cubic",
    "out_cubic",
    "in_out_cubic",
    "out_in_cubic",
    "in_quart",
    "out_quart",
    "in_out_quart",
    "out_in_quart",
    "in_quint",
    "out_quint",
    "in_out_quint",
    "out_in_quint",
    "in_sine",
    "out_sine",
    "in_out_sine",
    "out_in_sine",
    "in_expo",
    "out_expo",
    "in_out_expo",
    "out_in_expo",
    "in_circ",
    "out_circ",
    "in_out_circ",
    "out_in_circ",
    "in_elastic",
    "out_elastic",
    "in_out_elastic",
    "out_in_elastic",
    "in_back",
    "out_back",
    "in_out_back",
    "out_in_back",
    "in_bounce",
    "out_bounce",
    "in_out_bounce",
    "out_in_bounce",
    "bezier_spline",
    "tcb_spline",
    "custom",
]

TYPE: bidict[TypeStr, QtCore.QEasingCurve.Type] = bidict(
    linear=QtCore.QEasingCurve.Type.Linear,
    in_quad=QtCore.QEasingCurve.Type.InQuad,
    out_quad=QtCore.QEasingCurve.Type.OutQuad,
    in_out_quad=QtCore.QEasingCurve.Type.InOutQuad,
    out_in_quad=QtCore.QEasingCurve.Type.OutInQuad,
    in_cubic=QtCore.QEasingCurve.Type.InCubic,
    out_cubic=QtCore.QEasingCurve.Type.OutCubic,
    in_out_cubic=QtCore.QEasingCurve.Type.InOutCubic,
    out_in_cubic=QtCore.QEasingCurve.Type.OutInCubic,
    in_quart=QtCore.QEasingCurve.Type.InQuart,
    out_quart=QtCore.QEasingCurve.Type.OutQuart,
    in_out_quart=QtCore.QEasingCurve.Type.InOutQuart,
    out_in_quart=QtCore.QEasingCurve.Type.OutInQuart,
    in_quint=QtCore.QEasingCurve.Type.InQuint,
    out_quint=QtCore.QEasingCurve.Type.OutQuint,
    in_out_quint=QtCore.QEasingCurve.Type.InOutQuint,
    out_in_quint=QtCore.QEasingCurve.Type.OutInQuint,
    in_sine=QtCore.QEasingCurve.Type.InSine,
    out_sine=QtCore.QEasingCurve.Type.OutSine,
    in_out_sine=QtCore.QEasingCurve.Type.InOutSine,
    out_in_sine=QtCore.QEasingCurve.Type.OutInSine,
    in_expo=QtCore.QEasingCurve.Type.InExpo,
    out_expo=QtCore.QEasingCurve.Type.OutExpo,
    in_out_expo=QtCore.QEasingCurve.Type.InOutExpo,
    out_in_expo=QtCore.QEasingCurve.Type.OutInExpo,
    in_circ=QtCore.QEasingCurve.Type.InCirc,
    out_circ=QtCore.QEasingCurve.Type.OutCirc,
    in_out_circ=QtCore.QEasingCurve.Type.InOutCirc,
    out_in_circ=QtCore.QEasingCurve.Type.OutInCirc,
    in_elastic=QtCore.QEasingCurve.Type.InElastic,
    out_elastic=QtCore.QEasingCurve.Type.OutElastic,
    in_out_elastic=QtCore.QEasingCurve.Type.InOutElastic,
    out_in_elastic=QtCore.QEasingCurve.Type.OutInElastic,
    in_back=QtCore.QEasingCurve.Type.InBack,
    out_back=QtCore.QEasingCurve.Type.OutBack,
    in_out_back=QtCore.QEasingCurve.Type.InOutBack,
    out_in_back=QtCore.QEasingCurve.Type.OutInBack,
    in_bounce=QtCore.QEasingCurve.Type.InBounce,
    out_bounce=QtCore.QEasingCurve.Type.OutBounce,
    in_out_bounce=QtCore.QEasingCurve.Type.InOutBounce,
    out_in_bounce=QtCore.QEasingCurve.Type.OutInBounce,
    bezier_spline=QtCore.QEasingCurve.Type.BezierSpline,
    tcb_spline=QtCore.QEasingCurve.Type.TCBSpline,
    custom=QtCore.QEasingCurve.Type.Custom,
)


CurveMethod = Callable[[float], float]


class EasingCurve(serializemixin.SerializeMixin, QtCore.QEasingCurve):
    """Easing curves for controlling animation."""

    def __init__(self, other_or_type: TypeStr | int | QtCore.QEasingCurve = "linear"):
        if isinstance(other_or_type, str) and other_or_type in TYPE:
            typ = TYPE[other_or_type]
        else:
            typ = other_or_type
        super().__init__(typ)

    def __getitem__(self, value: float) -> float:
        return self.valueForProgress(value)

    def __repr__(self):
        return get_repr(self, self.get_type())

    def set_custom_type(self, method: CurveMethod):
        self.setCustomType(method)

    def get_custom_type(self) -> CurveMethod:
        return self.customType()  # type: ignore

    def set_type(self, typ: TypeStr | QtCore.QEasingCurve.Type):
        """Set easing curve type.

        Args:
            typ: easing curve type
        """
        self.setType(TYPE.get_enum_value(typ))

    def get_type(self) -> TypeStr:
        """Get the current easing curve type.

        Returns:
            easing curve type
        """
        return TYPE.inverse[self.type()]


if __name__ == "__main__":
    c = EasingCurve()
    print(repr(c))



from __future__ import annotations

from abc import ABCMeta
from collections.abc import Sequence
import datetime
import enum
import os
import pathlib
import re

from typing import TYPE_CHECKING, Any, ClassVar, Protocol, TypeVar, runtime_checkable
from urllib import parse

import dateutil.parser

from prettyqt.qt import QtCore


class QABCMeta(type(QtCore.QObject), ABCMeta):
    """Metaclass, mainly used for inheriting from MutableMapping."""


class Singleton(type(QtCore.QObject), type):
    """Metaclass to create a singleton."""

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kw)
        return cls.instance


Indexer = tuple[slice | int, slice | int]
IndexerOrInt = Indexer | int
SingleResultIndexer = int | tuple[int, int]
MultiResultIndexer = tuple[slice, int] | tuple[int, slice] | tuple[slice, slice]

PatternType = re.Pattern | QtCore.QRegularExpression
PatternAndStringType = str | PatternType
JSONType = str | int | float | bool | None | dict[str, Any] | list[Any]
PathType = str | os.PathLike


class Validatable(Protocol):
    """An object with an isValid method (e.g. QUrl)."""

    def isValid(self) -> bool:
        ...


class SupportsValue(Protocol):
    """An object with a set_value and get_value method."""

    def set_value(self, value):
        ...

    def get_value(self, value):
        ...


@runtime_checkable
class IsDataclass(Protocol):
    __dataclass_fields__: ClassVar[dict]


@runtime_checkable
class IsAttrs(Protocol):
    __attrs_attrs__: ClassVar[dict]


@runtime_checkable
class IsTreeIterator(Protocol):
    @property
    def root(self):
        ...

    @property
    def _abc_impl(self):
        ...


if TYPE_CHECKING:
    from prettyqt import core
    from prettyqt.qt import QtGui, QtWidgets

    UrlType = str | QtCore.QUrl
    PointType = tuple[int, int] | QtCore.QPoint
    PointFType = tuple[int | float, int | float] | QtCore.QPointF
    SizeType = tuple[int, int] | QtCore.QSize | int
    SizeFType = tuple[int | float, int | float] | QtCore.QSizeF
    MarginsType = tuple[int, int, int, int] | tuple[int, int] | int | QtCore.QMargins
    MarginsFType = (
        tuple[int | float, int | float, int | float, int | float]
        | tuple[int | float, int | float]
        | int
        | float
        | QtCore.QMarginsF
    )
    RectType = tuple[int, int, int, int] | QtCore.QRect
    RectFType = tuple[int | float, int | float, int | float, int | float] | QtCore.QRectF
    Vector3DType = (
        tuple[int | float, int | float, int | float]
        | QtCore.QPoint
        | QtCore.QPointF
        | QtGui.QVector2D
        | QtGui.QVector3D
        | QtGui.QVector4D
    )
    LineType = tuple[int, int] | QtCore.QLine
    LineFType = tuple[float, float] | QtCore.QLineF | QtCore.QLine
    SemanticVersionType = str | QtCore.QVersionNumber | tuple[int, int, int]
    IconType = QtGui.QIcon | str | pathlib.Path | None
    ByteArrayType = str | bytes | QtCore.QByteArray
    TimeType = QtCore.QTime | datetime.time | str | tuple[int, int, int]
    DateType = QtCore.QDate | datetime.date | str | tuple[int, int, int]
    DateTimeType = (
        QtCore.QDateTime | datetime.datetime | str | tuple[int, int, int, int, int, int]
    )
    KeySequenceType = QtGui.QKeySequence | QtCore.QKeyCombination | str | None
    # ContainerType = Union[widgets.Layout, widgets.Splitter]
    TransformType = (
        QtGui.QTransform
        | tuple[float, float, float, float, float, float, float, float, float]
        | tuple[float, float, float, float, float, float]
    )
    ColorType = (
        str
        | int
        | QtCore.Qt.GlobalColor
        | QtGui.QColor
        | tuple[int, int, int]
        | tuple[int, int, int, int]
        | None
    )
    KeyCombinationType = (
        str | QtCore.QKeyCombination | QtGui.QKeySequence | QtGui.QKeySequence.StandardKey
    )
    ColorAndBrushType = ColorType | QtGui.QBrush

    VariantType = (
        QtCore.QPersistentModelIndex
        | QtCore.QModelIndex
        | QtCore.QJsonDocument
        |
        # QtCore.QJsonArray |
        # QtCore.QJsonObject |
        QtCore.QJsonValue
        | QtCore.QUrl
        | QtCore.QUuid
        | QtCore.QEasingCurve
        | QtCore.QRegularExpression
        | QtCore.QLocale
        | QtCore.QRectF
        | QtCore.QRect
        | QtCore.QLineF
        | QtCore.QLine
        | QtCore.QPointF
        | QtCore.QPoint
        | QtCore.QSizeF
        | QtCore.QSize
        | QtCore.QDateTime
        | QtCore.QTime
        | QtCore.QDate
        | QtCore.QByteArray
        | QtGui.QColor
        | QtGui.QImage
        | QtGui.QBrush
        | QtGui.QBitmap
        | QtGui.QCursor
        | QtGui.QFont
        | QtGui.QKeySequence
        | QtGui.QIcon
        | QtGui.QTransform
        | QtGui.QPalette
        | QtGui.QMatrix4x4
        | QtGui.QPen
        | QtGui.QPixmap
        | QtGui.QPolygon
        | QtGui.QPolygonF
        | QtGui.QRegion
        | QtWidgets.QSizePolicy
        | QtGui.QTextFormat
        | QtGui.QTextLength
        | QtGui.QVector2D
        | QtGui.QVector3D
        | QtGui.QVector4D
        | bool
        | str
        | int
        | float
        | bytes
    )

    Variant = VariantType | list[VariantType] | dict[str, VariantType]

    QtSerializableType = (
        QtCore.QByteArray
        | QtCore.QUrl
        | QtGui.QBrush
        | QtGui.QColor
        | QtGui.QCursor
        | QtGui.QIcon
        | QtGui.QIconEngine
        | QtGui.QImage
        | QtGui.QPalette
        | QtGui.QPen
        | QtGui.QPicture
        | QtGui.QPixmap
        | QtGui.QPolygon
        | QtGui.QPolygonF
        | QtGui.QRegion
        | QtGui.QStandardItem
        | QtGui.QTransform
        | QtWidgets.QListWidgetItem
        | QtWidgets.QTreeWidgetItem
    )
    # QtGui.QColorSpace |
    # QtWebEngineCore.QWebEngineHistory


def to_string(val: Any, locale: QtCore.QLocale | None = None) -> str:
    from prettyqt import constants, core, gui, widgets

    if locale is None:
        locale = core.Locale()
    match val:
        case str():
            return val
        case bool():
            return "✓" if val else "☐"
        case enum.Flag():
            return val.name
        case enum.Enum():
            return val.name
        case int() | float() | core.QByteArray():
            return locale.toString(val)
        case gui.QColor():
            return f"({val.red()},{val.green()},{val.blue()},{val.alpha()})"
        case gui.QBrush():
            # TODO: can also be gradient
            val = val.color()
            return f"({val.red()},{val.green()},{val.blue()},{val.alpha()})"
        case gui.QFont():
            return val.family()
        case gui.QRegion():
            rect = val.boundingRect()
            return f"({rect.x()},{rect.y()},{rect.width()},{rect.height()})"
        case gui.QCursor():
            return constants.CURSOR_SHAPE.inverse[val.shape()]
        case gui.QKeySequence():
            return val.toString()
        case core.QDate() | core.QDateTime() | core.QTime():
            return val.toString(constants.DateFormat.ISODate)
        case core.QPoint():
            return f"({val.x()},{val.y()})"
        case core.QRect():
            return f"({val.x()},{val.y()},{val.width()},{val.height()})"
        case core.QSize():
            return f"({val.width()},{val.height()})"
        case core.QLocale():
            return val.bcp47Name()
        case widgets.QSizePolicy():
            return (
                f"({widgets.sizepolicy.SIZE_POLICY.inverse[val.horizontalPolicy()]}, "
                f"{widgets.sizepolicy.SIZE_POLICY.inverse[val.verticalPolicy()]}, "
                f"{widgets.sizepolicy.CONTROL_TYPE.inverse[val.controlType()]})"
            )
        case list():
            return ",".join(map(repr, val))
        case re.Pattern():
            return val.pattern
        case datetime.date():
            return val.isoformat()
        case datetime.datetime():
            return val.isoformat(sep=" ")
        case core.QRegularExpression():
            return val.pattern()
        case core.QUrl():
            return val.toString()
        case slice() | range():
            return f"({val.start=}, {val.stop=}, {val.step=}"
    try:
        import numpy as np
    except ImportError:
        pass
    else:
        match val:
            case np.integer():
                return to_string(int(val), locale)
            case np.floating():
                return to_string(float(val), locale)
            case np.str_():
                return val.astype(str)
            case np.bool_():
                return bool(val.astype(bool))
            case np.datetime64():
                return to_string(val.astype(datetime.datetime), locale)
    return repr(val)


def get_editor_for_value_list(ls: Sequence, parent=None):
    from prettyqt import widgets

    container = widgets.Widget(parent)
    container.set_layout("vertical")
    editors = [get_editor_for_value(val, parent=container) for val in ls]
    container.box.add(editors)
    container.set_window_flags("tool")
    return container


def get_editor_for_value(val, parent=None):
    """Returns an editor for given value.

    Functions checks for type and returns an appropriate editor.
    The returned editor is guaranteed to have a get_value and set_value method
    which can be used to the given value.

    Arguments:
        val: Any value type
        parent: parent widget for editor
    """
    from prettyqt import core, custom_widgets, gui, widgets

    match val:
        case bool():
            return widgets.CheckBox(parent=parent)
        case enum.Flag():
            widget = custom_widgets.EnumFlagWidget(parent=parent)
            widget._set_enum_class(type(val))
            return widget
        case enum.Enum():
            widget = custom_widgets.EnumComboBox(parent=parent)
            widget._set_enum_class(type(val))
            return widget
        case int():
            return widgets.SpinBox(parent=parent)
        case float():
            return widgets.DoubleSpinBox(parent=parent)
        case (int(), *_):
            return custom_widgets.ListInput(parent=parent, typ=int)
        case (float(), *_):
            return custom_widgets.ListInput(parent=parent, typ=float)
        case (str(), *_):
            return custom_widgets.StringListEdit(parent=parent)
        case pathlib.Path():
            return custom_widgets.FileChooserButton(parent=parent)
        case str():
            widget = widgets.LineEdit(parent=parent)
            widget.setFrame(False)
            return widget
        case core.QRegularExpression() | re.Pattern():
            return custom_widgets.RegexInput(show_error=False, parent=parent)
        case core.QTime():
            return widgets.TimeEdit(parent=parent)
        case core.QDate():
            return widgets.DateEdit(parent=parent)
        case core.QDateTime():
            return widgets.DateTimeEdit(parent=parent)
        case core.QPoint():
            return custom_widgets.PointEdit(parent=parent)
        case core.QSize():
            return custom_widgets.SizeEdit(parent=parent)
        case core.QRect():
            return custom_widgets.RectEdit(parent=parent)
        case gui.QKeySequence():
            return widgets.KeySequenceEdit(parent=parent)
        case gui.QRegion():
            return custom_widgets.RegionEdit(parent=parent)
        case gui.QFont():
            return widgets.FontComboBox(parent=parent)
        case gui.QColor():
            return custom_widgets.ColorComboBox(parent=parent)
        case gui.QBrush():
            return custom_widgets.BrushEdit(parent=parent)
        case widgets.QSizePolicy():
            return custom_widgets.SizePolicyEdit(parent=parent)
        case core.QUrl():
            return custom_widgets.UrlLineEdit(parent=parent)
        case gui.QPalette():
            return custom_widgets.PaletteEdit(parent=parent)
        case gui.QCursor():
            return custom_widgets.CursorEdit(parent=parent)
        case gui.QIcon():
            return custom_widgets.IconEdit(parent=parent)
        case core.QLocale():
            return custom_widgets.LocaleEdit(parent=parent)
        case slice():
            return custom_widgets.SliceEdit(parent=parent)
        case range():
            return custom_widgets.RangeEdit(parent=parent)
        # case QtCore.QRectF():  # todo
        #     return custom_widgets.RectEdit(parent=parent)
    try:
        import numpy as np
    except ImportError:
        pass
    else:
        match val:
            case np.floating():
                return custom_widgets.FloatLineEdit(parent=parent)
            case np.integer():
                return custom_widgets.IntLineEdit(parent=parent)
            case np.str_():
                return widgets.LineEdit(parent=parent)
            case np.datetime64():
                return widgets.DateTimeEdit(parent=parent)
            case np.bool_():
                return widgets.CheckBox(parent=parent)


T = TypeVar("T")


def align_types(source: T, target: VariantType | tuple) -> T:
    """Align target to the type of source."""
    from prettyqt import core, gui
    from prettyqt.utils import colors

    match source:
        case core.QPoint():
            return to_point(target)
        case core.QPointF():
            return to_pointf(target)
        case core.QSize():
            return to_size(target)
        case core.QSizeF():
            return to_sizef(target)
        case core.QRect():
            return to_rect(target)
        case core.QRectF():
            return to_rectf(target)
        case gui.QColor():
            return colors.get_color(target).as_qt()
        case _:
            return target


def make_qtype(obj):
    """Cast a subclassed PrettyQt instance to its orginal Qt Type."""
    from prettyqt import core

    match obj:
        case core.Margins():
            return QtCore.QMargins(obj)
        case core.Locale():
            return QtCore.QLocale(obj)
        case core.KeyCombination():
            return QtCore.QKeyCombination(obj)
        case core.Url():
            return QtCore.QUrl(obj)
        case core.EasingCurve():
            return QtCore.QEasingCurve(obj)
        case list():
            return [make_qtype(i) for i in obj]

    from prettyqt import gui
    from prettyqt.qt import QtGui

    match obj:
        case gui.Palette():
            return QtGui.QPalette(obj)
        case gui.Font():
            return QtGui.QFont(obj)
        case gui.Color():
            return obj.convertTo(obj.spec())
        case gui.Cursor():
            return QtGui.QCursor(obj)
        case gui.Brush():
            return QtGui.QBrush(obj)
        case gui.Pixmap():
            return QtGui.QPixmap(obj)
        case gui.Region():
            return QtGui.QRegion(obj)
        case gui.KeySequence():
            return QtGui.QKeySequence(obj)
        case gui.Icon():
            return QtGui.QIcon(obj)
        case gui.Vector3D():
            # PyQt doesnt allow Vector3D in ctor
            return QtGui.QVector3D(obj.x(), obj.y(), obj.z())
        case gui.Vector4D():
            # PyQt doesnt allow Vector4D in ctor
            return QtGui.QVector4D(obj.x(), obj.y(), obj.z(), obj.w())
    return obj


def to_point(point: QtCore.QPointF | PointType):
    match point:
        case (int(), int()):
            return QtCore.QPoint(*point)
        case QtCore.QPoint():
            return point
        case QtCore.QPointF():
            return point.toPointF()
        case _:
            raise TypeError(point)


def to_pointf(point: PointFType | QtCore.QPoint):
    match point:
        case (int() | float(), int() | float()):
            return QtCore.QPointF(*point)
        case QtCore.QPointF():
            return point
        case QtCore.QPoint():
            return point.toPointF()
        case _:
            raise TypeError(point)


def to_size(size: QtCore.QSizeF | SizeType):
    match size:
        case (int(), int()):
            return QtCore.QSize(*size)
        case int():
            return QtCore.QSize(size, size)
        case QtCore.QSize():
            return size
        case QtCore.QSizeF():
            return size.toSize()
        case _:
            raise TypeError(size)


def to_sizef(size: SizeFType | QtCore.QSize):
    match size:
        case (int() | float(), int() | float()):
            return QtCore.QSizeF(*size)
        case QtCore.QSizeF():
            return size
        case QtCore.QSize():
            return size.toSizeF()
        case _:
            raise TypeError(size)


def to_rect(rect: QtCore.QRectF | RectType):
    match rect:
        case (int(), int(), int(), int()):
            return QtCore.QRect(*rect)
        case QtCore.QRect():
            return rect
        case QtCore.QRectF():
            return rect.toRect()
        case _:
            raise TypeError(rect)


def to_rectf(rect: RectFType | QtCore.QRect):
    match rect:
        case (int() | float(), int() | float(), int() | float(), int() | float()):
            return QtCore.QRectF(*rect)
        case QtCore.QRectF():
            return rect
        case QtCore.QRect():
            return rect.toRectF()
        case _:
            raise TypeError(rect)


def to_vector3d(vector: Vector3DType):
    from prettyqt import core, gui

    match vector:
        case (int() | float(), int() | float(), int() | float()):
            return gui.QVector3D(*vector)
        case gui.QVector3D():
            return vector
        case gui.QVector2D() | gui.QVector4D() | core.QPoint() | core.QPointF():
            return gui.QVector3D(vector)
        case _:
            raise TypeError(vector)


def to_url(url: UrlType | None) -> QtCore.QUrl:
    match url:
        case os.PathLike():
            return QtCore.QUrl.fromLocalFile(os.fspath(url))
        case str():
            return QtCore.QUrl(url)
        case None:
            return QtCore.QUrl()
        case QtCore.QUrl():
            return url
        case parse.ParseResult():
            return QtCore.QUrl(url.geturl())
        case _:
            raise TypeError(url)


def to_local_url(url: UrlType | os.PathLike | None) -> QtCore.QUrl:
    # TODO: need to check whether we should merge to_local_url and to_url
    # core.Url.from_user_input() perhaps a good option, too?
    match url:
        case os.PathLike():
            return QtCore.QUrl.fromLocalFile(os.fspath(url))
        case str():
            return QtCore.QUrl.fromLocalFile(url)
        case None:
            return QtCore.QUrl()
        case QtCore.QUrl():
            return url
        case _:
            raise TypeError(url)


def to_bytearray(arr: str | bytes | QtCore.QByteArray) -> QtCore.QByteArray:
    match arr:
        case str():
            return QtCore.QByteArray(arr.encode())
        case bytes():
            return QtCore.QByteArray(arr)
        case QtCore.QByteArray():
            return arr
        case _:
            raise TypeError(arr)


def to_margins(margins: MarginsType | QtCore.QMarginsF | None) -> QtCore.QMargins:
    match margins:
        case (int() as x, int() as y):
            return QtCore.QMargins(x, y, x, y)
        case (int() as left, int() as top, int() as right, int() as bottom):
            return QtCore.QMargins(left, top, right, bottom)
        case QtCore.QMargins():
            return margins
        case QtCore.QMarginsF():
            return margins.toMargins()
        case int() as x:
            return QtCore.QMargins(x, x, x, x)
        case None:
            return QtCore.QMargins(0, 0, 0, 0)
        case _:
            raise TypeError(margins)


def to_marginsf(margins: MarginsFType | QtCore.QMargins | None) -> QtCore.QMarginsF:
    match margins:
        case (int() | float() as x, int() | float() as y):
            return QtCore.QMarginsF(x, y, x, y)
        case (
            int() | float() as left,
            int() | float() as top,
            int() | float() as right,
            int() | float() as bottom,
        ):
            return QtCore.QMarginsF(left, top, right, bottom)
        case QtCore.QMarginsF():
            return margins
        case QtCore.QMargins():
            return margins.toMarginsF()
        case int() | float() as x:
            return QtCore.QMarginsF(x, x, x, x)
        case None:
            return QtCore.QMarginsF(0, 0, 0, 0)
        case _:
            raise TypeError(margins)


def to_datetime(date_time: DateTimeType):
    match date_time:
        case None:
            return QtCore.QDateTime()
        case (int(), int(), int(), int(), int(), int()):
            date_ = QtCore.QDate(*date_time[:3])
            time_ = QtCore.QTime(*date_time[3:])
            return QtCore.QDateTime(date_, time_)
        case str():
            return dateutil.parser.parse(date_time)
        case QtCore.QDateTime() | datetime.datetime():
            return date_time
        case _:
            raise TypeError(date_time)


def to_date(value: DateType):
    match value:
        case None:
            return QtCore.QDate()
        case (int(), int(), int()):
            return QtCore.QDate(*value)
        case str():
            date = dateutil.parser.parse(value).date()
            return QtCore.QDate(date)
        case QtCore.QDate() | datetime.date():
            return value
        case _:
            raise TypeError(value)


def to_time(value: TimeType):
    from prettyqt import core

    match value:
        case str():
            val = core.Time.from_string(value)
            if not val.isValid():
                raise ValueError(value)
            return val
        case (int(), int(), int()):
            return QtCore.QTime(*value)
        case core.QTime():
            return core.Time(value)
        case core.QDateTime():
            return core.DateTime(value).get_time()
        case datetime.time():
            return core.Time(value)
        case datetime.datetime():
            return core.DateTime(value).get_time()
        case _:
            raise TypeError(value)


def to_linef(line: LineFType):
    match line:
        case QtCore.QLine():
            return QtCore.QLineF(line)
        case tuple():
            return QtCore.QLineF(*line)
        case QtCore.QLineF():
            return line
        case _:
            raise TypeError(line)


def to_py_pattern(pattern: PatternAndStringType):
    from prettyqt import core

    match pattern:
        case str():
            return re.compile(pattern)
        case re.Pattern():
            return pattern
        case QtCore.QRegularExpression():
            return core.RegularExpression(pattern).to_py_pattern()
        case _:
            raise TypeError(pattern)


def to_regular_expression(
    pattern: PatternAndStringType,
    flag: core.regularexpression.PatternOptionStr
    | core.RegularExpression.PatternOption
    | None = None,
):
    from prettyqt import core

    if flag is None:
        flag = core.RegularExpression.PatternOption.NoPatternOption
    else:
        flag = core.regularexpression.PATTERN_OPTIONS.get_enum_value(flag)
    match pattern:
        case str():
            return core.RegularExpression(pattern, flag)
        case re.Pattern():
            return core.RegularExpression(pattern)
        case QtCore.QRegularExpression():
            return core.RegularExpression(pattern)
        case _:
            raise TypeError(pattern)


def to_transform(transform: TransformType):
    match transform:
        case tuple():
            return QtGui.QTransform(*transform)
        case QtGui.QTransform():
            return transform
        case _:
            raise TypeError(transform)


def to_keysequence(sequence: KeySequenceType):
    from prettyqt import gui

    match sequence:
        case None:
            return gui.KeySequence("")
        case QtCore.QKeyCombination():
            return gui.KeySequence(sequence)
        case gui.QKeySequence():
            return sequence
        case str():
            return gui.KeySequence(sequence, gui.KeySequence.SequenceFormat.PortableText)
        case _:
            raise TypeError(sequence)


def make_serializable(obj):
    #  possible to avoid importing by checking the metaobject instead of isinstance?
    from prettyqt import core

    match obj:
        case QtCore.QMargins():
            return core.Margins(obj)
        case QtCore.QLocale():
            return core.Locale(obj)
        case QtCore.QKeyCombination():
            return core.KeyCombination(obj)
        case QtCore.QUrl():
            return core.Url(obj)
        case QtCore.QEasingCurve():
            return core.EasingCurve(obj)
        case list():
            return [make_serializable(i) for i in obj]

    from prettyqt import gui
    from prettyqt.qt import QtGui

    match obj:
        case QtGui.QPalette():
            return gui.Palette(obj)
        case QtGui.QFont():
            return gui.Font(obj)
        case QtGui.QCursor():
            return gui.Cursor(obj)
        case QtGui.QBrush():
            return gui.Brush(obj)
        case QtGui.QPixmap():
            return gui.Pixmap(obj)
        case QtGui.QRegion():
            return gui.Region(obj)
        case QtGui.QKeySequence():
            return gui.KeySequence(obj)
        case QtGui.QIcon():
            return gui.Icon(obj)
        case QtGui.QVector3D():
            # PyQt doesnt allow Vector3D in ctor
            return gui.Vector3D(obj.x(), obj.y(), obj.z())
        case QtGui.QVector4D():
            # PyQt doesnt allow Vector4D in ctor
            return gui.Vector4D(obj.x(), obj.y(), obj.z(), obj.w())
        case QtGui.QTextDocument():
            # TODO: cant serialize this yet
            return None

    from prettyqt import widgets
    from prettyqt.qt import QtWidgets

    match obj:
        case QtWidgets.QSizePolicy():
            return widgets.SizePolicy.clone(obj)
        case _:
            return obj



from __future__ import annotations

from prettyqt import gui, widgets


class FontDialog(widgets.DialogMixin, widgets.QFontDialog):
    """Dialog widget for selecting a font."""

    def get_current_font(self) -> gui.Font:
        return gui.Font(self.currentFont())


if __name__ == "__main__":
    app = widgets.app()
    widget = FontDialog.getFont()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui
from prettyqt.qt import QtCharts
from prettyqt.utils import bidict, get_repr


ValuePositionStr = Literal[
    "lower_extreme", "lower_quartile", "median", "upper_quartile", "upper_extreme"
]

VALUE_POSITION: bidict[ValuePositionStr, QtCharts.QBoxSet.ValuePositions] = bidict(
    lower_extreme=QtCharts.QBoxSet.ValuePositions.LowerExtreme,
    lower_quartile=QtCharts.QBoxSet.ValuePositions.LowerQuartile,
    median=QtCharts.QBoxSet.ValuePositions.Median,
    upper_quartile=QtCharts.QBoxSet.ValuePositions.UpperQuartile,
    upper_extreme=QtCharts.QBoxSet.ValuePositions.UpperExtreme,
)


class BoxSet(core.ObjectMixin, QtCharts.QBoxSet):
    def __repr__(self):
        return get_repr(
            self,
            self["lower_extreme"],
            self["lower_quartile"],
            self["median"],
            self["upper_quartile"],
            self["upper_extreme"],
            self.label(),
        )

    def __getitem__(
        self, index: int | ValuePositionStr | QtCharts.QBoxSet.ValuePositions
    ) -> float:
        """Returns the value of the box-and-whiskers item specified by index."""
        if type(index) is not int:
            index = VALUE_POSITION.get_enum_value(index).value
        if not (0 <= index <= 4):
            raise KeyError(index)
        return self.at(index)

    def __setitem__(
        self, index: int | ValuePositionStr | QtCharts.QBoxSet.ValuePositions, value: int
    ):
        """Sets the value specified by value in the position specified by index."""
        if type(index) is not int:
            index = VALUE_POSITION.get_enum_value(index).value
        if not (0 <= index <= 4):
            raise KeyError(index)
        self.setValue(index, value)

    def get_pen(self) -> gui.Pen:
        return gui.Pen(self.pen())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())


if __name__ == "__main__":
    boxset = BoxSet()
    boxset[0]



from __future__ import annotations

from math import hypot

from prettyqt import core, gui, widgets


class JoystickButton(widgets.PushButton):
    state_changed = core.Signal(object)

    def __init__(self, parent: widgets.QWidget | None = None):
        super().__init__(parent)
        self.radius = 200
        self.setCheckable(True)
        self._state = (0.0, 0.0)
        self.set_state((0, 0))
        self.setFixedWidth(50)
        self.setFixedHeight(50)

    def mousePressEvent(self, ev):
        self.setChecked(True)
        self.press_pos = ev.position()
        ev.accept()

    def mouseMoveEvent(self, ev):
        dif = ev.position() - self.press_pos
        self.set_state((dif.x(), -dif.y()))

    def mouseReleaseEvent(self, ev):
        self.setChecked(False)
        self.set_state((0, 0))

    def wheelEvent(self, ev):
        ev.accept()

    def doubleClickEvent(self, ev):
        ev.accept()

    def get_state(self) -> tuple[float, float]:
        return self._state

    def set_state(self, state: tuple[float, float]):
        d = hypot(state[0], state[1])  # length
        nxy = [0 if i == 0 else i / d for i in state]
        d = min(d, self.radius)
        d = (d / self.radius) ** 2
        state = (nxy[0] * d, nxy[1] * d)

        w2 = self.width() / 2
        h2 = self.height() / 2
        self.spot_pos = core.Point(int(w2 * (1 + state[0])), int(h2 * (1 - state[1])))
        self.update()
        if self._state == state:
            return
        self._state = state
        self.state_changed.emit(self._state)

    def paintEvent(self, ev):
        super().paintEvent(ev)
        p = gui.QPainter(self)
        p.setBrush(gui.QBrush(gui.QColor(0, 0, 0)))
        p.drawEllipse(self.spot_pos.x() - 3, self.spot_pos.y() - 3, 6, 6)

    def resizeEvent(self, ev):
        self.set_state(self._state)
        super().resizeEvent(ev)

    state = core.Property(
        tuple,
        get_state,
        set_state,
        doc="Button state.",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = JoystickButton()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import qthelp


class HelpEngine(qthelp.HelpEngineCoreMixin, qthelp.QHelpEngine):
    """Access to contents and indices of the help engine."""

    def get_content_model(self) -> qthelp.HelpContentModel:
        return qthelp.HelpContentModel(self.contentModel())


if __name__ == "__main__":
    engine = HelpEngine("")
    engine.get_files("a", "b")



from __future__ import annotations

from collections.abc import Iterable
import os
import pathlib

from prettyqt import gui, widgets
from prettyqt.utils import datatypes


class TextBrowser(widgets.TextEditMixin, widgets.QTextBrowser):
    """Rich text browser with hypertext navigation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setOpenExternalLinks(True)

    # def dragEnterEvent(self, event):
    #     u = event.mimeData().urls()
    #     for url in u:
    #         file_path = os.path.abspath(url.toLocalFile())

    #         ext = file_path.split(".")[-1]
    #         if ext in ["txt", "md", "markdown"]:
    #             event.accept()
    #         else:
    #             event.ignore()

    # def dropEvent(self, event):
    #     event.accept()
    #     self.show_markdown_file(self.filePath)

    def set_markdown_file(self, file_path: datatypes.PathType):
        file_path = pathlib.Path(file_path)
        with file_path.open() as f:
            file_content = f.read()
        self.setMarkdown(file_content)

    def get_search_paths(self) -> list[pathlib.Path]:
        return [pathlib.Path(p) for p in self.searchPaths()]

    def set_search_paths(self, paths: Iterable[datatypes.PathType]):
        self.setSearchPaths([os.fspath(p) for p in paths])

    def get_source_type(self) -> gui.textdocument.ResourceTypeStr:
        return gui.textdocument.RESOURCE_TYPES.inverse[self.sourceType()]


if __name__ == "__main__":
    app = widgets.app()
    reader = TextBrowser()
    reader.show()
    app.exec()



"""Module containing helper functions."""

from collections.abc import Iterable
import enum
import functools
import logging
import operator

from typing import Any, TypeVar

import bidict as bdct


logger = logging.getLogger(__name__)

KT = TypeVar("KT", bound=str)
VT = TypeVar("VT", bound=enum.Enum)


class bidict(bdct.bidict[KT, VT]):
    def __init__(self, *args, **kwargs):
        match args:
            case (dict(),):
                super().__init__(args[0])
            case _:
                super().__init__(kwargs)

    def __getitem__(self, item) -> VT:
        try:
            return super().__getitem__(item)
        except KeyError as e:
            raise InvalidParamError(item, list(self.keys())) from e

    def get_list(self, flag: enum.Enum) -> list[KT]:
        return [k for k, v in self.items() if v & flag]

    def get_dict(self, flag: enum.Enum) -> dict[KT, VT]:
        return {k: v & flag for k, v in self.items()}

    # def get_flag(self, **kwargs) -> dict[str, Any]:
    #     for k, v in kwargs.items():
    #     if isinstance(flag, enum.Enum):
    #         flag = flag.value
    #     return {k: v & flag for k, v in self.items()}

    def merge_flags(self, flags: Iterable[KT]) -> VT:
        return functools.reduce(operator.ior, [self[t] for t in flags])

    def get_enum_value(self, value: KT | VT) -> VT:
        return self[value] if isinstance(value, str) else value

    def get_str_value(self, value: KT | VT) -> KT:
        return value if isinstance(value, str) else self.inverse[value]


class InvalidParamError(ValueError):
    """Exception raised for invalid params in method calls.

    Args:
        value: param value which caused the error
        valid_options: allowed options
    """

    def __init__(self, value: Any, valid_options: Iterable):
        self.value = value
        opts = " / ".join(repr(opt) for opt in valid_options)
        self.message = f"Invalid value: {value!r}. Allowed options are {opts}."
        super().__init__(self.message)


def get_repr(_obj: Any, *args: Any, **kwargs: Any) -> str:
    """Get a suitable __repr__ string for an object.

    Args:
        _obj: The object to get a repr for.
        *args: Arguments for the repr
        **kwargs: Keyword arguments for the repr
    """
    classname = type(_obj).__name__
    parts = [repr(val) for val in args]
    kw_parts = [f"{name}={val!r}" for name, val in kwargs.items()]
    return f"{classname}({', '.join(parts + kw_parts)})"



from __future__ import annotations

from typing import Any

from prettyqt import constants, core


class ModelMixin:
    DTYPE_ROLE = constants.USER_ROLE + 1  # type: ignore
    HEADER = ["Name"]
    LABELS: dict = {}
    CHECKSTATE: dict = {}
    TOOLTIPS: dict = {}
    DECORATIONS: dict = {}
    SET_DATA: dict = {}
    content_type = ""

    def headerData(
        self,
        offset: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self.HEADER[offset]

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        """Override for AbstractitemModel base method.

        returns corresponding flags for cell of supplied index
        """
        if not index.isValid():
            return constants.DROP_ENABLED
        if index.column() in self.SET_DATA:
            return self.DEFAULT_FLAGS | constants.IS_EDITABLE
        return self.DEFAULT_FLAGS

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        item = self.data_by_index(index)
        match role:
            case constants.DECORATION_ROLE:
                if fn := self.DECORATIONS.get(index.column()):
                    return fn(item)
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                if fn := self.LABELS.get(index.column()):
                    return fn(item)
            case constants.TOOLTIP_ROLE:
                if fn := self.TOOLTIPS.get(index.column()):
                    return fn(item)
            case constants.CHECKSTATE_ROLE:
                if fn := self.CHECKSTATE.get(index.column()):
                    return fn(item)
            case constants.USER_ROLE:
                return item
            case _:
                return None

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if role == constants.EDIT_ROLE:
            if not value:
                return False
            item = self.data_by_index(index)
            if fn := self.SET_DATA.get(index.column()):
                fn(item, value)
                self.update_row(index.row())
                return True



from __future__ import annotations

import os

from prettyqt import constants, core
from prettyqt.qt import QtSvg
from prettyqt.utils import datatypes


class SvgRenderer(core.ObjectMixin, QtSvg.QSvgRenderer):
    def load_file(self, path: datatypes.PathType):
        result = self.load(os.fspath(path))
        if not result:
            raise ValueError("invalid path")

    def set_aspect_ratio_mode(
        self, mode: constants.AspectRatioModeStr | constants.AspectRatioMode
    ):
        """Set the aspect ratio mode.

        Args:
            mode: aspect ratio mode
        """
        self.setAspectRatioMode(constants.ASPECT_RATIO_MODE.get_enum_value(mode))

    def get_aspect_ratio_mode(self) -> constants.AspectRatioModeStr:
        """Return current aspect ratio mode.

        Returns:
            aspect ratio mode
        """
        return constants.ASPECT_RATIO_MODE.inverse[self.aspectRatioMode()]



from __future__ import annotations

import collections

from collections.abc import Iterator, Sequence
import contextlib
import functools
import importlib
import inspect
import logging
import operator
import types
import typing


T = typing.TypeVar("T", bound=type)


logger = logging.getLogger(__name__)

# def add_docs(klass):
#     import pathlib
#     from prettyqt import paths
#     klass_name = klass.__name__
#     qclass = klass
#     for k in klass.mro():
#         if k.__name__.startswith("Q"):
#             qclass = k
#             break
#     module = k.__module__.split(".")[1]
#     path = paths.DOCSTRING_PATH /  module
#     filepath = path / f"Q{klass_name.replace('Mixin', '')}.txt"
#     if filepath.exists():
#         klass.__doc__ =  filepath.read_text()
#     return klass


def lca_type(classes: list[type]) -> type:
    return next(
        iter(
            functools.reduce(
                operator.and_,
                (collections.Counter(Klass.mro()) for Klass in classes),
            )
        )
    )


def to_module(
    module: str | Sequence[str] | types.ModuleType,
    return_none: bool = True,
) -> types.ModuleType | None:
    """Returns a module for given module path. If module is given, just return it.

    Arguments:
        module: A ModuleType, "module.submodule[...]" or ["module", "submodule", [...]]
        return_none: In case module cant get imported, return None.
                     If False, Exception is thrown.
    """
    match module:
        case (str(), *_) | str():
            module_path = module if isinstance(module, str) else ".".join(module)
            try:
                return importlib.import_module(module_path)
            except (ImportError, AttributeError) as e:
                logger.warning(f"Could not import {module_path!r}")
                if return_none:
                    return
                raise e
        case types.ModuleType():
            return module
        case _:
            raise TypeError(module)


def to_module_parts(module: Sequence[str] | str | types.ModuleType) -> tuple[str, ...]:
    """Returns a tuple describing the module path.

    Result is of form ("module", "submodule", "subsubmodule")

    Arguments:
        module: A ModuleType, str or sequence.
    """
    match module:
        case (str(), *_):
            return tuple(module)
        case str():
            return tuple(module.split("."))
        case types.ModuleType():
            return tuple(module.__name__.split("."))
        case _:
            raise TypeError(module)


def find_common_ancestor(cls_list: list[type]) -> type:
    mros = [list(inspect.getmro(cls)) for cls in cls_list]
    track = collections.defaultdict(int)
    while mros:
        for mro in mros:
            cur = mro.pop(0)
            track[cur] += 1
            if (
                track[cur] >= len(cls_list)
                and not cur.__name__.endswith("Mixin")
                and cur.__name__.startswith("Q")
            ):
                return cur
            if len(mro) == 0:
                mros.remove(mro)
    raise TypeError("Couldnt find common base class")


def get_subclasses(klass: type, include_abstract: bool = False) -> typing.Iterator[type]:
    """Recursively iter all subclasses of given klass.

    Arguments:
        klass: class to get subclasses from
        include_abstract: whether abstract base classes should be included.
    """
    if getattr(klass.__subclasses__, "__self__", None) is None:
        return
    for i in klass.__subclasses__():
        yield from get_subclasses(i)
        if include_abstract or not inspect.isabstract(i):
            yield i


def get_class_by_name(
    klass_name: str, *, parent_type: type = object, module_filter: list[str] | None = None
) -> type | None:
    """Get class by given Name.

    Arguments:
        klass_name: class name to search for
        parent_type: Only check subclasses of given type.
        module_filter: Only consider classes which do not include any of the strs in the
                      module path.
    """
    return next(
        (
            kls
            for kls in get_subclasses(parent_type)
            if kls.__name__ == klass_name
            and (module_filter is None or kls.__module__.split(".")[0] in module_filter)
        ),
        None,
    )


def get_qt_parent_class(klass: type, only_direct: bool = False) -> type | None:
    """Get Qt-based parent class for given type.

    Arguments:
        klass: class to get parent type from
        only_direct: whether only the direct parent should be checked.
    """
    if only_direct:
        name = klass.mro()[0].__name__
        return klass if name.startswith(("PyQt", "PySide")) else None
    return next(
        (kls for kls in klass.mro() if kls.__module__.startswith(("PyQt", "PySide"))),
        None,
    )


def get_class_for_id(base_class: T, id_: str) -> T:
    """Get subclass of base class which has an attribute "ID" with given id.

    Arguments:
        base_class: Class to check subclasses from.
        id_: id to search for.
    """
    base_classes = (
        typing.get_args(base_class)
        if isinstance(base_class, types.UnionType)
        else (base_class,)
    )
    for base_class in base_classes:
        for Klass in get_subclasses(base_class):
            if "ID" in Klass.__dict__ and id_ == Klass.ID:
                logger.debug(f"found class for id {Klass.ID!r}")
                return Klass
    raise ValueError(f"Couldnt find class with id {id_!r} for base class {base_class}")


def iter_classes_for_module(
    module: types.ModuleType | str | tuple[str],
    *,
    type_filter: type | None | types.UnionType = None,
    module_filter: str | None = None,
    filter_by___all__: bool = False,
    recursive: bool = False,
) -> Iterator[type]:
    """Yield classes from given module.

    Arguments:
        module: either a module or a path to a module in form of str or
                tuple of strings.
        type_filter: only yield classes which are subclasses of given type.
        module_filter: filter by a module prefix.
        filter_by___all__: Whether to filter based on whats defined in __all__.
        recursive: import all submodules recursively and also yield their classes.
    """
    module = to_module(module)
    if not module:
        return []
    if recursive:
        for _name, submod in inspect.getmembers(module, inspect.ismodule):
            if submod.__name__.startswith(module_filter or ""):
                yield from iter_classes_for_module(
                    submod,
                    type_filter=type_filter,
                    module_filter=submod.__name__,
                    filter_by___all__=filter_by___all__,
                    recursive=True,
                )
    for klass_name, kls in inspect.getmembers(module, inspect.isclass):
        has_all = hasattr(module, "__all__")
        if filter_by___all__ and (not has_all or klass_name not in module.__all__):
            continue
        if type_filter is not None and not issubclass(kls, type_filter):
            continue
        if module_filter is not None and not kls.__module__.startswith(module_filter):
            continue
        yield kls


def get_topmost_module_path_for_klass(klass: type) -> str:
    """Return path of topmost module containing given class.

    If a class is imported in any of its parent modules, return that "shorter" path.

    Arguments:
        klass: Klass to get the path for.
    """
    path = klass.__module__
    parts = path.split(".")
    while parts:
        with contextlib.suppress(TypeError):
            new_path = ".".join(parts)
            mod = importlib.import_module(new_path)
            klasses = [kls for _kls_name, kls in inspect.getmembers(mod, inspect.isclass)]
            if klass in klasses:
                path = new_path
        parts = parts[:-1]
    return path


if __name__ == "__main__":
    from prettyqt import widgets

    path = iter_classes_for_module(widgets, recursive=False)
    print(len(list(path)))



from __future__ import annotations

from prettyqt import widgets


class WidgetItem(widgets.LayoutItemMixin, widgets.QWidgetItem):
    """Layout item that represents a widget."""


if __name__ == "__main__":
    app = widgets.app()
    item = WidgetItem(widgets.QWidget())



from __future__ import annotations

from prettyqt import core, eventfilters


class TextUpdateEventFilter(eventfilters.BaseEventFilter):
    def __init__(self, parent, interval: int = 1000, **kwargs):
        super().__init__(parent, **kwargs)
        parent.startTimer(interval)
        self._update_text()

    def eventFilter(self, obj, event: core.QEvent) -> bool:
        # if obj is not self.parent():
        #     return super().eventFilter(obj, event)
        match event.type():
            case core.Event.Type.Timer:
                self._update_text()
                return True
        return super().eventFilter(obj, event)

    def _update_text(self):
        return NotImplemented


class TimeLabelEventFilter(TextUpdateEventFilter):
    ID = "time_label"

    def __init__(self, parent, time_format: str = "hh:mm:ss", **kwargs):
        self._format = time_format
        super().__init__(parent, **kwargs)

    def _update_text(self):
        time = core.Time.get_current_time()
        self.parent().set_text(time.toString(self._format))


class DateLabelEventFilter(TextUpdateEventFilter):
    ID = "date_label"

    def __init__(self, parent, date_format: str = "dd.MM.yyyy", **kwargs):
        self._format = date_format
        super().__init__(parent, **kwargs)

    def _update_text(self):
        time = core.Date.get_current_date()
        self.parent().set_text(time.toString(self._format))


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.PushButton()
    widget.set_icon("mdi.folder")
    test = DateLabelEventFilter(widget)
    widget.installEventFilter(test)
    widget.show()
    app.exec()



from __future__ import annotations

from .alphanumericvalidator import AlphaNumericValidator
from .blacklistvalidator import BlacklistValidator, NotEmptyValidator, NotZeroValidator
from .colorvalidator import ColorValidator
from .compositevalidator import AndValidator, OrValidator
from .functionvalidator import FunctionValidator
from .hexvalidator import HexValidator
from .integervalidator import IntegerValidator
from .jsonvalidator import JsonValidator
from .monotoniclistvalidator import MonotonicListValidator
from .notstrictvalidator import NotStrictValidator
from .pathvalidator import PathValidator
from .pythoncodevalidator import PythonCodeValidator
from .qssvalidator import QssValidator
from .regexpatternvalidator import RegexPatternValidator
from .regexvalidators import FloatListValidator, IntListValidator
from .scientificvalidators import ScientificFloatValidator, ScientificIntegerValidator
from .textlengthvalidator import TextLengthValidator
from .whitelistvalidator import EmptyValidator, WhitelistValidator


__all__ = [
    "PathValidator",
    "FunctionValidator",
    "NotEmptyValidator",
    "IntegerValidator",
    "IntListValidator",
    "FloatListValidator",
    "NotZeroValidator",
    "AndValidator",
    "OrValidator",
    "RegexPatternValidator",
    "HexValidator",
    "QssValidator",
    "TextLengthValidator",
    "WhitelistValidator",
    "EmptyValidator",
    "BlacklistValidator",
    "ScientificIntegerValidator",
    "ScientificFloatValidator",
    "PythonCodeValidator",
    "JsonValidator",
    "ColorValidator",
    "AlphaNumericValidator",
    "NotStrictValidator",
    "MonotonicListValidator",
]



# credits to SuperQt (https://github.com/pyapp-kit/superqt)

from __future__ import annotations

import enum

from typing import TypeVar

from prettyqt import core, widgets
from prettyqt.utils import get_repr


EnumType = TypeVar("EnumType", bound=enum.Enum)


NONE_STRING = "----"


class EnumComboBox(widgets.ComboBox):
    """ComboBox presenting options from a python Enum.

    If the Enum class does not implement `__str__` then a human readable name
    is created from the name of the enum member, replacing underscores with spaces.
    """

    # current_enum_changed = core.Signal(object)
    value_changed = core.Signal(enum.Enum)

    def __init__(self, value=None, object_name: str = "enum_combobox", **kwargs):
        self._enum_class = None
        self._allow_none = False
        super().__init__(object_name=object_name, **kwargs)
        if value is not None:
            self.set_value(value)
        # self.currentIndexChanged.connect(self._emit_signal)

    def __repr__(self):
        return get_repr(self, self.get_value())

    def set_allow_none(self, value: bool):
        self._allow_none = value

    def is_none_allowed(self) -> bool:
        return self._allow_none

    def _set_enum_class(self, enum: enum.EnumMeta | None):
        """Set enum class from which members value should be selected."""
        if enum == self._enum_class:
            return None
        self._enum_class = enum
        super().clear()
        if self._allow_none and enum is not None:
            super().addItem(NONE_STRING)
        items = [i.name.replace("_", " ") for i in self._enum_class.__members__.values()]
        super().addItems(items)

    def get_enum_class(self) -> enum.EnumMeta | None:
        """Return current Enum class."""
        return self._enum_class

    def is_optional(self) -> bool:
        """Return if current enum is with optional annotation."""
        return self._allow_none

    def clear(self):
        self._enum_class = None
        self._allow_none = False
        super().clear()

    def get_value(self) -> EnumType | None:
        """Current value as Enum member."""
        if self._enum_class is None:
            return None
        class_members = list(self._enum_class.__members__.values())
        if not self._allow_none:
            return class_members[self.currentIndex()]
        is_none = self.currentText() == NONE_STRING
        return None if is_none else class_members[self.currentIndex() - 1]

    def set_value(self, value: EnumType | None) -> None:
        """Set value with Enum."""
        if value is None:
            if not self._allow_none:
                raise ValueError(value)
            self.setCurrentIndex(0)
            return
        if not isinstance(value, enum.Enum):
            value = self._enum_class(value)
        self._set_enum_class(value.__class__)
        self.setCurrentText(value.name.replace("_", " "))

    # def _emit_signal(self):
    #     if self._enum_class is not None:
    #         self.current_enum_changed.emit(self.get_value())

    allowNone = core.Property(
        bool,
        is_none_allowed,
        set_allow_none,
        doc="Whether None is allowed as a value",
    )
    enumValue = core.Property(
        enum.Enum,
        get_value,
        set_value,
        user=True,
        doc="Currently chosen value",
    )
    # enumClass = core.Property(type(enum.Enum), get_enum_class, set_enum_class)


if __name__ == "__main__":
    from prettyqt import constants

    app = widgets.app()
    cb = EnumComboBox(allow_none=False)
    cb.set_value(constants.ItemDataRole.EditRole)
    cb.show()
    cb.value_changed.connect(print)
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtPdf
from prettyqt.utils import bidict


ERROR = bidict(
    none=QtPdf.QPdfDocument.Error.None_,
    unknown=QtPdf.QPdfDocument.Error.Unknown,
    data_not_yet_available=QtPdf.QPdfDocument.Error.DataNotYetAvailable,
    file_not_found=QtPdf.QPdfDocument.Error.FileNotFound,
    invalid_file_format=QtPdf.QPdfDocument.Error.InvalidFileFormat,
    incorrect_password=QtPdf.QPdfDocument.Error.IncorrectPassword,
    unsupported_security_scheme=QtPdf.QPdfDocument.Error.UnsupportedSecurityScheme,
)

ErrorStr = Literal[
    "none",
    "unknown",
    "data_not_yet_available",
    "file_not_found",
    "invalid_file_format",
    "incorrect_password",
    "unsupported_security_scheme",
]

STATUS = bidict(
    null=QtPdf.QPdfDocument.Status.Null,
    loading=QtPdf.QPdfDocument.Status.Loading,
    ready=QtPdf.QPdfDocument.Status.Ready,
    unloading=QtPdf.QPdfDocument.Status.Unloading,
    error=QtPdf.QPdfDocument.Status.Error,
)

StatusStr = Literal[
    "null",
    "loading",
    "ready",
    "unloading",
    "error",
]


class PdfDocument(core.ObjectMixin, QtPdf.QPdfDocument):
    """Loads a PDF document and renders pages from it."""

    def get_error(self) -> ErrorStr:
        """Return current error.

        Returns:
            page mode
        """
        return ERROR.inverse[self.error()]

    def get_status(self) -> StatusStr:
        """Return current status.

        Returns:
            zoom mode
        """
        return STATUS.inverse[self.status()]


if __name__ == "__main__":
    doc = PdfDocument()
    print(doc.get_status())



from __future__ import annotations

from typing import Literal

from prettyqt import core, network
from prettyqt.utils import bidict


SOCKET_OPTION = bidict(
    none=network.QLocalServer.SocketOption(0),
    user=network.QLocalServer.SocketOption.UserAccessOption,
    group=network.QLocalServer.SocketOption.GroupAccessOption,
    other=network.QLocalServer.SocketOption.OtherAccessOption,
    world=network.QLocalServer.SocketOption.WorldAccessOption,
)

SocketOptionStr = Literal["none", "user", "group", "other", "world"]


class LocalServer(core.ObjectMixin, network.QLocalServer):
    """Local socket based server."""

    def get_server_error(self) -> network.abstractsocket.SocketErrorStr:
        return network.abstractsocket.SOCKET_ERROR.inverse[self.serverError()]

    def set_socket_options(self, *name: SocketOptionStr):
        flags = SOCKET_OPTION.merge_flags(name)
        self.setSocketOptions(flags)

    def get_socket_options(self) -> list[SocketOptionStr]:
        return SOCKET_OPTION.get_list(self.socketOptions())


if __name__ == "__main__":
    server = LocalServer()



from __future__ import annotations

from collections.abc import MutableMapping
from typing import Literal

from prettyqt.qt import QtBluetooth
from prettyqt.utils import bidict, datatypes


AttributeIdStr = Literal[
    "service_record_handle",
    "service_class_ids",
    "service_record_state",
    "service_id",
    "protocol_descriptor_list",
    "browse_group_list",
    "language_base_attribute_id_list",
    "service_info_time_to_live",
    "service_availablity",
    "bluetooth_profile_descriptor_list",
    "documentation_url",
    "client_executable_url",
    "icon_url",
    "additional_protocol_descriptor_list",
    "primary_language_base",
    "service_name",
    "service_description",
    "service_provider",
]

AttributeId = QtBluetooth.QBluetoothServiceInfo.AttributeId

ATTRIBUTE_IDS: bidict[AttributeIdStr, AttributeId] = bidict(
    service_record_handle=AttributeId.ServiceRecordHandle,
    service_class_ids=AttributeId.ServiceClassIds,
    service_record_state=AttributeId.ServiceRecordState,
    service_id=AttributeId.ServiceId,
    protocol_descriptor_list=AttributeId.ProtocolDescriptorList,
    browse_group_list=AttributeId.BrowseGroupList,
    language_base_attribute_id_list=AttributeId.LanguageBaseAttributeIdList,
    service_info_time_to_live=AttributeId.ServiceInfoTimeToLive,
    service_availablity=AttributeId.ServiceAvailability,
    bluetooth_profile_descriptor_list=AttributeId.BluetoothProfileDescriptorList,
    documentation_url=AttributeId.DocumentationUrl,
    client_executable_url=AttributeId.ClientExecutableUrl,
    icon_url=AttributeId.IconUrl,
    additional_protocol_descriptor_list=AttributeId.AdditionalProtocolDescriptorList,
    primary_language_base=AttributeId.PrimaryLanguageBase,
    # service_name=AttributeId.ServiceName,
    service_description=AttributeId.ServiceDescription,
    service_provider=AttributeId.ServiceProvider,
)

Protocol = QtBluetooth.QBluetoothServiceInfo.Protocol

ProtocolStr = Literal["unknown", "l2_cap", "rfcomm"]

PROTOCOL: bidict[ProtocolStr, Protocol] = bidict(
    unknown=Protocol.UnknownProtocol,
    l2_cap=Protocol.L2capProtocol,
    rfcomm=Protocol.RfcommProtocol,
)


class BluetoothServiceInfo(
    QtBluetooth.QBluetoothServiceInfo, MutableMapping, metaclass=datatypes.QABCMeta
):
    """Enables access to the attributes of a Bluetooth service.

    Also implements MutableMapping interface, can be used as a dicionary.
    """

    def __getitem__(self, value: str | int | AttributeId):
        match value:
            case int():
                flag = value
            case str():
                if value not in ATTRIBUTE_IDS:
                    raise KeyError(value)
                flag = ATTRIBUTE_IDS[value].value
            case AttributeId():
                flag = value.value
            case _:
                raise KeyError(value)
        return self.attribute(flag)

    def __delitem__(self, value: str | int | AttributeId):
        match value:
            case int():
                flag = value
            case str():
                flag = ATTRIBUTE_IDS[value].value
            case AttributeId():
                flag = value.value
        return self.removeAttribute(flag)

    def __setitem__(self, index: str | int | AttributeId, value):
        """Set attribute."""
        match index:
            case int():
                flag = index
            case str():
                flag = ATTRIBUTE_IDS[index].value
            case AttributeId():
                flag = index.value
        return self.setAttribute(flag, value)

    def __contains__(self, value: int) -> bool:
        attr = ATTRIBUTE_IDS.inverse[value]
        return self.contains(attr)

    def __iter__(self):
        """Iter the info attributes."""
        return iter(self.attributes())

    def __len__(self):
        return len(self.attributes())


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    address = BluetoothServiceInfo()
    address["documentation_url"] = "test"
    del address["documentation_url"]



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.utils import bidict, datatypes


ShadowStr = Literal["plain", "raised", "sunken"]

SHADOW: bidict[ShadowStr, widgets.QFrame.Shadow] = bidict(
    plain=widgets.QFrame.Shadow.Plain,
    raised=widgets.QFrame.Shadow.Raised,
    sunken=widgets.QFrame.Shadow.Sunken,
)

FrameShapeStr = Literal[
    "no_frame", "box", "panel", "styled_panel", "h_line", "v_line", "win_panel"
]

FRAME_SHAPE: bidict[FrameShapeStr, widgets.QFrame.Shape] = bidict(
    no_frame=widgets.QFrame.Shape.NoFrame,
    box=widgets.QFrame.Shape.Box,
    panel=widgets.QFrame.Shape.Panel,
    styled_panel=widgets.QFrame.Shape.StyledPanel,
    h_line=widgets.QFrame.Shape.HLine,
    v_line=widgets.QFrame.Shape.VLine,
    win_panel=widgets.QFrame.Shape.WinPanel,
)


class FrameMixin(widgets.WidgetMixin):
    def _get_map(self):
        maps = super()._get_map()
        maps |= {"frameShape": FRAME_SHAPE, "frameShadow": SHADOW}
        return maps

    def set_frame_shadow(self, style: ShadowStr | widgets.QFrame.Shadow):
        """Set frame shadow.

        Args:
            style: frame style to use
        """
        self.setFrameShadow(SHADOW.get_enum_value(style))

    def get_frame_shadow(self) -> ShadowStr | None:
        """Return current frame shadow.

        Returns:
            frame style
        """
        if (frame_shadow := self.frameShadow()) == 0:
            return None
        return SHADOW.inverse[frame_shadow]

    def set_frame_shape(self, shape: FrameShapeStr | widgets.QFrame.Shape):
        """Set frame shape.

        Args:
            shape: frame shape to use
        """
        self.setFrameShape(FRAME_SHAPE.get_enum_value(shape))

    def get_frame_shape(self) -> FrameShapeStr:
        """Return current frame shape.

        Returns:
            frame shape
        """
        return FRAME_SHAPE.inverse[self.frameShape()]

    def set_frame_rect(self, rect: datatypes.RectType):
        self.setFrameRect(datatypes.to_rect(rect))


class Frame(FrameMixin, widgets.QFrame):
    """The base class of widgets that can have a frame."""


if __name__ == "__main__":
    app = widgets.app()
    widget = Frame(frame_shape="panel", object_name="fff")
    widget.show()
    app.exec()



"""Classes and functions for embedding web content in applications."""

import logging

logger = logging.getLogger(__name__)

try:
    from prettyqt.qt.QtWebEngineCore import *  # noqa: F403

    from .webenginehistoryitem import WebEngineHistoryItem
    from .webenginehistory import WebEngineHistory
    from .webenginehttprequest import WebEngineHttpRequest
    from .webenginecontextmenurequest import WebEngineContextMenuRequest
    from .webenginedownloadrequest import WebEngineDownloadRequest
    from .webengineurlscheme import WebEngineUrlScheme
    from .webengineurlschemehandler import WebEngineUrlSchemeHandler
    from .webenginesettings import WebEngineSettings
    from .webenginepage import WebEnginePage
    from .webenginescript import WebEngineScript
    from .webenginescriptcollection import WebEngineScriptCollection
    from .webengineprofile import WebEngineProfile
    from prettyqt.qt import QtWebEngineCore

    QT_MODULE = QtWebEngineCore
    __all__ = [
        "WebEngineHttpRequest",
        "WebEngineUrlScheme",
        "WebEngineHistoryItem",
        "WebEngineContextMenuRequest",
        "WebEngineDownloadRequest",
        "WebEngineHistory",
        "WebEngineUrlSchemeHandler",
        "WebEngineSettings",
        "WebEngineScript",
        "WebEngineScriptCollection",
        "WebEngineProfile",
        "WebEnginePage",
    ]
except ModuleNotFoundError:
    logger.warning("Tried to import WebEngine module but not installed.")



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict


RestartHintStr = Literal["if_running", "anyway", "immediately", "never"]

RESTART_HINT: bidict[RestartHintStr, QtGui.QSessionManager.RestartHint] = bidict(
    if_running=QtGui.QSessionManager.RestartHint.RestartIfRunning,
    anyway=QtGui.QSessionManager.RestartHint.RestartAnyway,
    immediately=QtGui.QSessionManager.RestartHint.RestartImmediately,
    never=QtGui.QSessionManager.RestartHint.RestartNever,
)


class SessionManager(core.ObjectMixin):
    def __init__(self, item: QtGui.QSessionManager):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def set_restart_hint(self, style: RestartHintStr | QtGui.QSessionManager.RestartHint):
        """Set the restart hint.

        Args:
            style: restart hint
        """
        self.setRestartHint(RESTART_HINT.get_enum_value(style))

    def get_restart_hint(self) -> RestartHintStr:
        """Return current restart hint.

        Returns:
            restart hint
        """
        return RESTART_HINT.inverse[self.restartHint()]



from __future__ import annotations

import enum
import logging

from typing import Any

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class EditorDelegate(widgets.StyledItemDelegate):
    """Delegate which supports editing a large range of data types.

    Allows editing a large range of different types from Qt, builtin libraries as well
    as some Numpy types.

    The Delegate also has an extended displayText method to show a repr-like string for
    different data types.

    The following data types are supported:

    * bool
    * enum.Flag
    * enum.Enum
    * int
    * float
    * str
    * range
    * slice
    * list of ints
    * list of floats
    * list of strings
    * pathlib.Path
    * re.Pattern
    * datetime.date
    * datetime.time
    * datetime.datetime
    * QtCore.QRegularExpression
    * QtCore.QTime
    * QtCore.QDate
    * QtCore.QDateTime
    * QtCore.QPoint
    * QtCore.QPointF
    * QtCore.QRect
    * QtCore.QRectF
    * QtCore.QRection
    * QtCore.QKeyCombination
    * QtCore.QLocale
    * QtCore.QSize
    * QtCore.QSizeF
    * QtCore.QUrl
    * QtGui.QFont
    * QtGui.QKeySequence
    * QtGui.QPalette
    * QtGui.QColor
    * QtGui.QBrush
    * QtGui.QCursor
    * QtGui.QIcon
    * QtWidgets.QSizePolicy

    If numpy is installed, the following types are supported, too:

    * numpy.floating
    * numpy.integer
    * numpy.str_
    * numpy.datetime64
    * numpy.bool_

    """

    ID = "editor"

    def __init__(
        self,
        *args,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
        validator: gui.QValidator
        | widgets.lineedit.ValidatorStr
        | datatypes.PatternType
        | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._role = role
        self._validator = validator

    def paint(
        self,
        painter: gui.QPainter,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        # if not self.is_supported_type(value):
        #     option = widgets.StyleOptionViewItem(option)
        #     option.state &= ~widgets.QStyle.StateFlag.State_Enabled
        from prettyqt import itemdelegates

        match value := self._data_for_index(index, self._role):
            case gui.QIcon():
                icon_delegate = itemdelegates.IconDelegate()
                icon_delegate.paint(painter, option, index)
                return
            case enum.Enum():  # PySide6 needs this when using Views
                option.text = value.name
                option.widget.style().drawControl(
                    widgets.QStyle.ControlElement.CE_ItemViewItem, option, painter
                )
            case _:
                super().paint(painter, option, index)

    def createEditor(
        self,
        parent: widgets.QWidget,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        val = self._data_for_index(index, self._role)
        logger.info(f"creating editor for {val!r}...")
        if isinstance(val, list):
            widget = datatypes.get_editor_for_value_list(val, parent)
        else:
            widget = datatypes.get_editor_for_value(val, parent)
        if widget is None:
            logger.warning(f"Could not find editor for {val!r} ({type(val)})")
            return None
        if self._validator and isinstance(
            widget, widgets.LineEdit | widgets.AbstractSpinBoxMixin
        ):
            widget.set_validator(self._validator, append=True)

        widget.setAutoFillBackground(True)
        widget.set_focus_policy("strong")
        return widget

    def setEditorData(self, editor: widgets.QWidget, index: core.ModelIndex):
        value = self._data_for_index(index, self._role)
        logger.info(f"setting data for {editor!r} to {value!r}")
        editor.set_value(value)

    def setModelData(
        self,
        editor: widgets.QWidget,
        model: core.QAbstractItemModel,
        index: core.ModelIndex,
    ):
        if (value := editor.get_value()) is not None:
            logger.info(f"setting data for {model!r} to {value!r}")
            model.setData(index, value, self._role)
            # self.closeEditor.emit(editor, self.EndEditHint.NoHint)
            self.commitData.emit(editor)

    def displayText(self, value: Any, locale: core.QLocale) -> str:
        return datatypes.to_string(value, locale)


if __name__ == "__main__":
    import pathlib
    import re

    app = widgets.app()
    table_widget = widgets.TableWidget(15, 2)
    table_widget.set_edit_triggers("all")
    table_widget.set_selection_behavior("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Rating"])
    types = dict(
        color=gui.QColor(30, 30, 30),
        time=core.QTime(1, 1, 1),
        date=core.QDate(1, 1, 1),
        datetime=core.QDateTime(1, 1, 1, 1, 1, 1),
        font=gui.QFont(),
        str="fdsf",
        int=8,
        float=8.2,
        path=pathlib.Path(),
        url=core.QUrl("http://www.google.de"),
        # regex=core.QRegularExpression("[a-z]"),
        regex=re.compile("[a-z]"),
        enum=constants.ItemDataRole.UserRole,
        bool=True,
        keysequence=gui.QKeySequence("Ctrl+A"),
    )
    for i, (k, v) in enumerate(types.items()):
        item_1 = widgets.TableWidgetItem(k)
        item_2 = widgets.TableWidgetItem()
        item_2.setData(constants.DISPLAY_ROLE, v)
        item_2.setData(constants.USER_ROLE, v)
        table_widget[i, 0] = item_1
        table_widget[i, 1] = item_2

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()
    with app.debug_mode():
        delegate = table_widget.set_delegate("editor", column=1)
        delegate.closeEditor.connect(print)
        app.exec()



from __future__ import annotations

from prettyqt.qt import QtLocation
from prettyqt.utils import get_repr


class PlaceContactDetail(QtLocation.QPlaceContactDetail):
    """Represents a contact detail such as a phone number or website url."""

    def __init__(
        self,
        other_or_label: None | str | QtLocation.QPlaceAttribute = None,
        value: str | None = None,
    ):
        if isinstance(other_or_label, QtLocation.QPlaceAttribute):
            super().__init__(other_or_label)
        else:
            super().__init__()
            self.setLabel(other_or_label or "")
            self.setValue(value or "")

    def __repr__(self):
        return get_repr(self, self.label(), self.value())

    def __str__(self):
        return f"{self.label()}: {self.value()}"


if __name__ == "__main__":
    detail = PlaceContactDetail("a", "b")



from __future__ import annotations

from prettyqt import constants, core


class AbstractTableModelMixin(core.AbstractItemModelMixin):
    DEFAULT_FLAGS = (
        constants.DRAG_ENABLED
        | constants.IS_ENABLED
        | constants.IS_SELECTABLE
        | constants.NO_CHILDREN
    )


class AbstractTableModel(AbstractTableModelMixin, core.QAbstractTableModel):
    """Abstract model that can be subclassed to create table models."""


if __name__ == "__main__":
    from prettyqt import itemmodels, widgets

    app = widgets.app()
    model = itemmodels.QObjectPropertiesModel(app)
    table = widgets.TableView()
    table.set_model(model)
    print(model.to_markdown())



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
import ctypes

from ctypes import WINFUNCTYPE, FormatError, c_bool, c_int
from ctypes.wintypes import UINT
import logging
from typing import SupportsInt

from prettyqt import core
from prettyqt.utils.platforms.windows import keytables


logger = logging.getLogger(__name__)


class GlobalHotKeyEventFilter(core.AbstractNativeEventFilter):
    _keybindings: defaultdict[list] = defaultdict(list)
    _keygrabs: defaultdict[int] = defaultdict(int)  # Key grab key -> number of grabs

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user32 = ctypes.WinDLL("user32", use_errno=True, use_last_error=True)
        # msdn.microsoft.com/en-us/library/windows/desktop/ms646309%28v=vs.85%29.aspx
        prototype = WINFUNCTYPE(c_bool, c_int, c_int, UINT, UINT)
        paramflags = (1, "hWnd", 0), (1, "id", 0), (1, "fsModifiers", 0), (1, "vk", 0)
        self.RegisterHotKey = prototype(("RegisterHotKey", self.user32), paramflags)

        # msdn.microsoft.com/en-us/library/windows/desktop/ms646327%28v=vs.85%29.aspx
        prototype = WINFUNCTYPE(c_bool, c_int, c_int)
        paramflags = (1, "hWnd", 0), (1, "id", 0)
        self.UnregisterHotKey = prototype(("UnregisterHotKey", self.user32), paramflags)

    def nativeEventFilter(self, eventType, message):
        WM_HOTKEY_MSG = 0x0312  # 768
        msg = ctypes.wintypes.MSG.from_address(int(message))
        handled = False
        if eventType != b"windows_generic_MSG" or msg.message != WM_HOTKEY_MSG:
            return False, 0
        key = msg.lParam
        logger.info(f"handling hotkey {key}")
        for cb in self._keybindings.get(key, []):
            try:
                cb()
            finally:
                handled = True
        return handled, 0

    def register_hotkey(
        self,
        keycombo: core.QKeyCombination,
        callback: Callable,
        wid: SupportsInt | None = None,
    ):
        wid = 0x0 if wid is None else int(wid)
        mods = keytables.qt_mod_to_virtual(keycombo.keyboardModifiers())
        kc = keytables.virtual_key_for_qtkey(keycombo.key())
        # High word = Key code, Low word = Modifiers
        # https://msdn.microsoft.com/en-us/library/windows/desktop/ms646279%28v=vs.85%29.aspx
        # Add MOD_NOREPEAT = 0x4000 to mods, so that keys don't get notified twice
        # This requires VISTA+ operating system
        key_index = kc << 16 | mods
        logger.info(f"registering key {key_index} for window {wid}")
        if not self._keygrabs[key_index] and not self.RegisterHotKey(
            wid, key_index, UINT(mods | 0x4000), UINT(kc)
        ):
            logger.warning("Couldn't register hot key!")
            return False

        self._keybindings[key_index].append(callback)
        self._keygrabs[key_index] += 1
        return True

    def unregister_hotkey(
        self, keycombo: core.QKeyCombination, wid: SupportsInt | None = None
    ):
        wid = 0x0 if wid is None else int(wid)
        mods = keytables.qt_mod_to_virtual(keycombo.keyboardModifiers())
        kc = keytables.virtual_key_for_qtkey(keycombo.key())
        key_index = kc << 16 | mods

        self._keybindings.pop(key_index)
        self._keygrabs.pop(key_index)

        logger.info(f"unregistering key {key_index} for window {wid}")
        if not self.UnregisterHotKey(wid, key_index):
            err = f"Couldn't unregister hot key '{kc}': {FormatError()}."
            logger.error(err)
            return False
        return True


if __name__ == "__main__":
    import sys

    from prettyqt import gui, widgets

    app = widgets.app()
    window = widgets.MainWindow()
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    logger.info("\tPress Ctrl+Shift+A or Print Screen any where to trigger a callback.")
    logger.info("\tCtrl+Shift+F unregisters and re-registers previous callback.")
    logger.info("\tCtrl+Shift+E exits the app.")

    # Setup a global keyboard shortcut to logger.info "Hello World" on pressing
    # the shortcut
    win_event_filter = GlobalHotKeyEventFilter()

    def callback():
        logger.info("hello world")

    def exit_app():
        window.close()

    def unregister():
        combo = gui.KeySequence("Shift+Ctrl+A")[0]
        win_event_filter.unregister_hotkey(combo)
        logger.info("unregister and register previous binding")
        win_event_filter.register_hotkey(combo, callback)

    combo = gui.KeySequence("Shift+Ctrl+A")[0]
    win_event_filter.register_hotkey(combo, callback)
    combo = gui.KeySequence("Print Screen")[0]
    win_event_filter.register_hotkey(combo, callback)
    combo = gui.KeySequence("Shift+Ctrl+E")[0]
    win_event_filter.register_hotkey(combo, exit_app)
    combo = gui.KeySequence("Shift+Ctrl+F")[0]
    win_event_filter.register_hotkey(combo, unregister)

    # # Install a native event filter to receive events from the OS
    win_event_filter.install()

    window.show()
    app.exec()
    combo = gui.KeySequence("Shift+Ctrl+A")[0]
    win_event_filter.unregister_hotkey(combo)
    combo = gui.KeySequence("Shift+Ctrl+F")[0]
    win_event_filter.unregister_hotkey(combo)
    combo = gui.KeySequence("Shift+Ctrl+E")[0]
    win_event_filter.unregister_hotkey(combo)



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtHelp


class HelpSearchQueryWidget(widgets.WidgetMixin, QtHelp.QHelpSearchQueryWidget):
    """Widget to enable the user to input a search term in a standardized input mask."""



from __future__ import annotations

from typing import Literal

from prettyqt import charts
from prettyqt.utils import bidict


LabelsPositionStr = Literal["center", "inside_end", "inside_base", "outside_end"]

LABELS_POSITIONS: bidict[LabelsPositionStr, charts.QAbstractBarSeries.LabelsPosition] = (
    bidict(
        center=charts.QAbstractBarSeries.LabelsPosition.LabelsCenter,
        inside_end=charts.QAbstractBarSeries.LabelsPosition.LabelsInsideEnd,
        inside_base=charts.QAbstractBarSeries.LabelsPosition.LabelsInsideBase,
        outside_end=charts.QAbstractBarSeries.LabelsPosition.LabelsOutsideEnd,
    )
)


class AbstractBarSeriesMixin(charts.AbstractSeriesMixin):
    def __delitem__(self, item: int | charts.QBarSet):
        """Implements `del series[0]`."""
        if isinstance(item, int):
            barsets = self.barSets()
            item = barsets[item]
        self.remove(item)

    def __getitem__(self, index: int) -> charts.QBarSet:
        """Implements `barset = series[1]`."""
        barsets = self.barSets()
        return barsets[index]

    def set_labels_position(
        self, position: LabelsPositionStr | charts.QAbstractBarSeries.LabelsPosition
    ):
        """Set the labels position.

        Args:
            position: Labels position
        """
        self.setLabelsPosition(LABELS_POSITIONS.get_enum_value(position))

    def get_labels_position(self) -> LabelsPositionStr:
        """Return current labels position.

        Returns:
            Labels position
        """
        return LABELS_POSITIONS.inverse[self.labelsPosition()]


class AbstractBarSeries(AbstractBarSeriesMixin, charts.QAbstractBarSeries):
    pass



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.qt import QtPrintSupport
from prettyqt.utils import bidict


ViewModeStr = Literal["single_page", "facing_pages", "all_pages"]

VIEW_MODE: bidict[ViewModeStr, QtPrintSupport.QPrintPreviewWidget.ViewMode] = bidict(
    single_page=QtPrintSupport.QPrintPreviewWidget.ViewMode.SinglePageView,
    facing_pages=QtPrintSupport.QPrintPreviewWidget.ViewMode.FacingPagesView,
    all_pages=QtPrintSupport.QPrintPreviewWidget.ViewMode.AllPagesView,
)

ZoomModeStr = Literal[
    "custom_zoom",
    "fit_to_width",
    "fit_in_view",
]

ZOOM_MODE: bidict[ZoomModeStr, QtPrintSupport.QPrintPreviewWidget.ZoomMode] = bidict(
    custom_zoom=QtPrintSupport.QPrintPreviewWidget.ZoomMode.CustomZoom,
    fit_to_width=QtPrintSupport.QPrintPreviewWidget.ZoomMode.FitToWidth,
    fit_in_view=QtPrintSupport.QPrintPreviewWidget.ZoomMode.FitInView,
)


class PrintPreviewWidget(widgets.WidgetMixin, QtPrintSupport.QPrintPreviewWidget):
    """Widget for previewing page layouts for printer output."""

    def get_view_mode(self) -> ViewModeStr:
        return VIEW_MODE.inverse[self.viewMode()]

    def set_view_mode(
        self, mode: ViewModeStr | QtPrintSupport.QPrintPreviewWidget.ViewMode
    ):
        """Set view mode.

        Args:
            mode: view mode
        """
        self.setViewMode(VIEW_MODE.get_enum_value(mode))

    def get_zoom_mode(self) -> ZoomModeStr:
        return ZOOM_MODE.inverse[self.zoomMode()]

    def set_zoom_mode(
        self, mode: ZoomModeStr | QtPrintSupport.QPrintPreviewWidget.ZoomMode
    ):
        """Set zoom mode.

        Args:
            mode: zoom mode
        """
        self.setZoomMode(ZOOM_MODE.get_enum_value(mode))



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Literal, TypedDict

from prettyqt import core
from prettyqt.qt import QtNetwork
from prettyqt.utils import bidict, datatypes, get_repr


Req = QtNetwork.QNetworkRequest
CE = Req.Attribute.ConnectionCacheExpiryTimeoutSecondsAttribute

AttributeStr = Literal[
    "http_status_code",
    "http_reason_phrase",
    "redirection_target",
    "connection_encrypted",
    "cache_load_control",
    "cache_save_control",
    "source_is_from_cache",
    "do_not_buffer_upload_data",
    "http_pipelining_allowed",
    "http_pipelining_was_used",
    "custom_verb",
    "cookie_load_control",
    "cookie_save_control",
    "authentication_reuse",
    "background_request",
    "http2_allowed",
    "http2_was_used",
    "emit_all_upload_progress_signals",
    "original_content_length",
    "redirect_policy",
    "http2_direct",
    "auto_delete_reply_on_finish",
    "connection_cache_expiry_timeout_seconds",
    "http2_cleartext_allowed",
    "use_credentials",
    "user",
    "user_max",
]

ATTRIBUTE: bidict[AttributeStr, Req.Attribute] = bidict(
    http_status_code=Req.Attribute.HttpStatusCodeAttribute,
    http_reason_phrase=Req.Attribute.HttpReasonPhraseAttribute,
    redirection_target=Req.Attribute.RedirectionTargetAttribute,
    connection_encrypted=Req.Attribute.ConnectionEncryptedAttribute,
    cache_load_control=Req.Attribute.CacheLoadControlAttribute,
    cache_save_control=Req.Attribute.CacheSaveControlAttribute,
    source_is_from_cache=Req.Attribute.SourceIsFromCacheAttribute,
    do_not_buffer_upload_data=Req.Attribute.DoNotBufferUploadDataAttribute,
    http_pipelining_allowed=Req.Attribute.HttpPipeliningAllowedAttribute,
    http_pipelining_was_used=Req.Attribute.HttpPipeliningWasUsedAttribute,
    custom_verb=Req.Attribute.CustomVerbAttribute,
    cookie_load_control=Req.Attribute.CookieLoadControlAttribute,
    cookie_save_control=Req.Attribute.CookieSaveControlAttribute,
    authentication_reuse=Req.Attribute.AuthenticationReuseAttribute,
    background_request=Req.Attribute.BackgroundRequestAttribute,
    http2_allowed=Req.Attribute.Http2AllowedAttribute,
    http2_was_used=Req.Attribute.Http2WasUsedAttribute,
    emit_all_upload_progress_signals=Req.Attribute.EmitAllUploadProgressSignalsAttribute,
    original_content_length=Req.Attribute.OriginalContentLengthAttribute,
    redirect_policy=Req.Attribute.RedirectPolicyAttribute,
    http2_direct=Req.Attribute.Http2DirectAttribute,
    auto_delete_reply_on_finish=Req.Attribute.AutoDeleteReplyOnFinishAttribute,
    connection_cache_expiry_timeout_seconds=CE,
    http2_cleartext_allowed=Req.Attribute.Http2CleartextAllowedAttribute,
    use_credentials=Req.Attribute.UseCredentialsAttribute,
    user=Req.Attribute.User,
    user_max=Req.Attribute.UserMax,
)


class TypedAttribute(TypedDict):
    http_status_code: int
    http_reason_phrase: bytes
    redirection_target: core.QUrl
    connection_encrypted: bool
    cache_load_control: int
    cache_save_control: bool
    source_is_from_cache: bool
    do_not_buffer_upload_data: bool
    http_pipelining_allowed: bool
    http_pipelining_was_used: bool
    custom_verb: bytes
    cookie_load_control: int
    cookie_save_control: int
    authentication_reuse: int
    background_request: bool
    http2_allowed: bool
    http2_was_used: bool
    emit_all_upload_progress_signals: bool
    original_content_length: int
    redirect_policy: int
    http2_direct: bool
    auto_delete_reply_on_finish: bool
    user: Any
    user_max: Any


KnownHeaderStr = Literal[
    "content_disposition",
    "content_type",
    "content_length",
    "location",
    "last_modified",
    "if_modified_since",
    "etag",
    "if_match",
    "if_none_match",
    "cookie",
    "set_cooke",
    "user_agent",
    "server",
]

KNOWN_HEADER: bidict[KnownHeaderStr, Req.KnownHeaders] = bidict(
    content_disposition=Req.KnownHeaders.ContentDispositionHeader,
    content_type=Req.KnownHeaders.ContentTypeHeader,
    content_length=Req.KnownHeaders.ContentLengthHeader,
    location=Req.KnownHeaders.LocationHeader,
    last_modified=Req.KnownHeaders.LastModifiedHeader,
    if_modified_since=Req.KnownHeaders.IfModifiedSinceHeader,
    etag=Req.KnownHeaders.ETagHeader,
    if_match=Req.KnownHeaders.IfMatchHeader,
    if_none_match=Req.KnownHeaders.IfNoneMatchHeader,
    cookie=Req.KnownHeaders.CookieHeader,
    set_cooke=Req.KnownHeaders.SetCookieHeader,
    user_agent=Req.KnownHeaders.UserAgentHeader,
    server=Req.KnownHeaders.ServerHeader,
)

PriorityStr = Literal["high", "normal", "low"]

PRIORITY: bidict[PriorityStr, Req.Priority] = bidict(
    high=Req.Priority.HighPriority,
    normal=Req.Priority.NormalPriority,
    low=Req.Priority.LowPriority,
)

RedirectPolicyStr = Literal["manual", "no_less_safe", "same_origin", "user_verified"]

REDIRECT_POLICIES: bidict[RedirectPolicyStr, Req.RedirectPolicy] = bidict(
    manual=Req.RedirectPolicy.ManualRedirectPolicy,
    no_less_safe=Req.RedirectPolicy.NoLessSafeRedirectPolicy,
    same_origin=Req.RedirectPolicy.SameOriginRedirectPolicy,
    user_verified=Req.RedirectPolicy.UserVerifiedRedirectPolicy,
)

CacheLoadControlStr = Literal[
    "always_network", "prefer_network", "prefer_cache", "always_cache"
]

CACHE_LOAD_CONTROL: bidict[CacheLoadControlStr, Req.CacheLoadControl] = bidict(
    always_network=Req.CacheLoadControl.AlwaysNetwork,
    prefer_network=Req.CacheLoadControl.PreferNetwork,
    prefer_cache=Req.CacheLoadControl.PreferCache,
    always_cache=Req.CacheLoadControl.AlwaysCache,
)
# else:
#     CACHE_LOAD_CONTROL = bidict(
#         always_network=Req.CacheLoadControlAlwaysNetwork,
#         prefer_network=Req.CacheLoadControlPreferNetwork,
#         prefer_cache=Req.CacheLoadControlPreferCache,
#         always_cache=Req.CacheLoadControlAlwaysCache,
# )


class NetworkRequest(QtNetwork.QNetworkRequest):
    """Holds a request to be sent with QNetworkAccessManager."""

    def __init__(self, obj: datatypes.UrlType | QtNetwork.QNetworkRequest | None = None):
        if isinstance(obj, QtNetwork.QNetworkRequest):
            super().__init__(obj)
        else:
            super().__init__()
            if obj is not None:
                self.set_url(obj)

    def __repr__(self):
        return get_repr(self, self.get_url())

    def set_header(self, name: KnownHeaderStr | Req.KnownHeaders, value: str):
        self.setHeader(KNOWN_HEADER.get_enum_value(name), value)

    def get_header(self, name: KnownHeaderStr | Req.KnownHeaders) -> str:
        return self.header(KNOWN_HEADER.get_enum_value(name))

    def set_headers(self, headers: Mapping[str, str]):
        for k, v in headers.items():
            self.setRawHeader(core.QByteArray(k.encode()), core.QByteArray(v.encode()))

    def get_headers(self) -> Mapping[str, str]:
        return {
            h.data().decode(): self.rawHeader(h).data().decode()
            for h in self.rawHeaderList()
        }

    def set_url(self, url: str | core.QUrl):
        url = core.Url(url)
        self.setUrl(url)

    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def set_priority(self, priority: PriorityStr | Req.Priority):
        """Set priority.

        Args:
            priority: priority
        """
        self.setPriority(PRIORITY.get_enum_value(priority))

    def get_priority(self) -> PriorityStr:
        """Get the current priority.

        Returns:
            priority
        """
        return PRIORITY.inverse[self.priority()]

    def set_attribute(
        self, attribute: AttributeStr | Req.Attribute, value: datatypes.Variant
    ):
        self.setAttribute(ATTRIBUTE.get_enum_value(attribute), value)

    def set_attributes(self, **kwargs):
        for k, v in kwargs.items():
            self.set_attribute(k, v)



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


class ButtonGroup(core.ObjectMixin, QtWidgets.QButtonGroup):
    """Container to organize groups of button widgets."""

    def __getitem__(self, index: int) -> QtWidgets.QAbstractButton:
        return self.button(index)



from __future__ import annotations

from collections.abc import Iterator

from prettyqt.qt import QtCore
from prettyqt.utils import get_repr


class Margins(QtCore.QMargins):
    """Defines the four margins of a rectangle."""

    def __repr__(self):
        return get_repr(self, self.left(), self.top(), self.right(), self.bottom())

    def __reduce__(self):
        return type(self), (self.left(), self.top(), self.right(), self.bottom())

    @property
    def _left(self) -> int:
        return self.left()

    @property
    def _top(self) -> int:
        return self.top()

    @property
    def _right(self) -> int:
        return self.right()

    @property
    def _bottom(self) -> int:
        return self.bottom()

    __match_args__ = ("_left", "_top", "_right", "_bottom")

    def __bool__(self):
        return not self.isNull()

    def __iter__(self) -> Iterator[int]:
        yield self.left()
        yield self.top()
        yield self.right()
        yield self.bottom()


if __name__ == "__main__":
    margins = Margins(0, 0, 0, 0)
    print(bool(margins))



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtHelp


class HelpIndexWidget(widgets.ListViewMixin, QtHelp.QHelpIndexWidget):
    """List view displaying the QHelpIndexModel."""

    @classmethod
    def setup_example(cls):
        return None



from __future__ import annotations

import logging

from typing import Literal

from prettyqt import core, custom_widgets, itemmodels, widgets


logger = logging.getLogger(__name__)


class ProxyComparerWidget(widgets.Splitter):
    """Splitter widget showing a comparison view between a proxy and its sourceModels.

    <figure markdown>
      ![Image title](../../images/proxycomparerwidget.png)
      <figcaption>ProxyComparerWidget</figcaption>
    </figure>
    """

    def __init__(
        self,
        model: core.QAbstractProxyModel,
        itemview: Literal["tree", "table", "list"]
        | type[widgets.QAbstractItemView] = "table",
        delegate: widgets.abstractitemview.DelegateStr
        | widgets.QItemDelegate
        | None = "editor",
        link_selections: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model_itemviews = []
        self.linker = None
        # determine ItemView class for the models
        match itemview:
            case "tree":
                View = widgets.TreeView
            case "table":
                View = widgets.TableView
            case "list":
                View = widgets.ListView
            case type():
                View = itemview
            case _:
                raise TypeError(itemview)

        # collect models
        models = []
        while isinstance(model, core.QAbstractProxyModel):
            models.append(model)
            model = model.sourceModel()
        models.append(model)

        # add column for each model

        for model in reversed(models):
            container = widgets.Widget()
            layout = container.set_layout("vertical")
            table = View()
            table.set_model(model)
            table.set_delegate(delegate)
            self.model_itemviews.append(table)
            prop_table = custom_widgets.QObjectPropertiesTableView()
            prop_table.set_qobject(model)
            header = widgets.Label(type(model).__name__)
            layout.add(header)
            col_splitter = widgets.Splitter("vertical")
            col_splitter.add(table)
            col_splitter.add(prop_table)
            layout.add(col_splitter)
            self.add(container)
        if link_selections:
            self.linker = itemmodels.LinkedSelectionModel(*self.model_itemviews)

    # @classmethod
    # def setup_example(cls):
    #     from prettyqt import gui

    #     dct = dict(
    #         a=["a", "a", "a", "a"],
    #         b=["b", "b", "b", "b"],
    #         c=["c", "c", "c", "c"],
    #     )
    #     model = gui.StandardItemModel.from_dict(dct)
    #     proxy = core.TransposeProxyModel(source_model=model)
    #     return cls(proxy)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    # table = debugging.example_table()
    # table.proxifier.transpose()
    # table.proxifier.to_list()
    splitter = ProxyComparerWidget.setup_example()
    splitter.show()
    app.exec()



from __future__ import annotations

import os

from prettyqt import widgets
from prettyqt.qt import QtSvgWidgets
from prettyqt.utils import datatypes


class SvgWidget(widgets.WidgetMixin, QtSvgWidgets.QSvgWidget):
    def load_file(self, path: datatypes.PathType):
        self.load(os.fspath(path))



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui, widgets
from prettyqt.utils import bidict, colors, datatypes


ViewModeStr = Literal["default", "tabbed"]

VIEW_MODE: bidict[ViewModeStr, widgets.QMdiArea.ViewMode] = bidict(
    default=widgets.QMdiArea.ViewMode.SubWindowView,
    tabbed=widgets.QMdiArea.ViewMode.TabbedView,
)

WindowOrderStr = Literal["creation", "stacking", "activation_history"]

WINDOW_ORDER: bidict[WindowOrderStr, widgets.QMdiArea.WindowOrder] = bidict(
    creation=widgets.QMdiArea.WindowOrder.CreationOrder,
    stacking=widgets.QMdiArea.WindowOrder.StackingOrder,
    activation_history=widgets.QMdiArea.WindowOrder.ActivationHistoryOrder,
)

TabPositionStr = Literal["north", "south", "west", "east"]

TAB_POSITION: bidict[TabPositionStr, widgets.QTabWidget.TabPosition] = bidict(
    north=widgets.QTabWidget.TabPosition.North,
    south=widgets.QTabWidget.TabPosition.South,
    west=widgets.QTabWidget.TabPosition.West,
    east=widgets.QTabWidget.TabPosition.East,
)


class MdiArea(widgets.AbstractScrollAreaMixin, widgets.QMdiArea):
    """Area in which MDI windows are displayed."""

    def __add__(self, other: widgets.QWidget):
        self.add(other)
        return self

    def set_view_mode(self, mode: ViewModeStr | widgets.QMdiArea.ViewMode):
        """Set view mode for the MDI area.

        Args:
            mode: view mode to use
        """
        self.setViewMode(VIEW_MODE.get_enum_value(mode))

    def get_view_mode(self) -> ViewModeStr:
        """Return current view mode.

        Returns:
            view mode
        """
        return VIEW_MODE.inverse[self.viewMode()]

    def set_window_order(self, mode: WindowOrderStr | widgets.QMdiArea.WindowOrder):
        """Set the window order behaviour for the MDI area.

        Args:
            mode: window order behaviour to use
        """
        self.setActivationOrder(WINDOW_ORDER.get_enum_value(mode))

    def get_window_order(self) -> WindowOrderStr:
        """Return current window order.

        Returns:
            view mode
        """
        return WINDOW_ORDER.inverse[self.activationOrder()]

    def set_tab_position(self, position: TabPositionStr | widgets.QTabWidget.TabPosition):
        """Set tab position for the MDI area.

        Args:
            position: tabs position to use
        """
        self.setTabPosition(TAB_POSITION.get_enum_value(position))

    def get_tab_position(self) -> TabPositionStr:
        """Return current tab position.

        Returns:
            tab position
        """
        return TAB_POSITION.inverse[self.tabPosition()]

    def set_background(
        self,
        brush_or_color: datatypes.ColorAndBrushType,
        pattern: constants.BrushStyleStr = "solid",
    ):
        if isinstance(brush_or_color, gui.QBrush):
            brush = brush_or_color
        else:
            color = colors.get_color(brush_or_color)
            brush = gui.Brush(color, constants.BRUSH_STYLE[pattern])
        self.setBackground(brush)

    def get_background(self) -> gui.Brush:
        return gui.Brush(self.background())

    def add(self, *item: widgets.QWidget):
        for i in item:
            self.add_subwindow(i)

    def add_subwindow(self, widget: widgets.QWidget) -> widgets.QMdiSubWindow:
        if not isinstance(widget, widgets.QMdiSubWindow):
            window = widgets.MdiSubWindow()
            window.setWidget(widget)
            self.addSubWindow(window)
            return window
        else:
            return self.addSubWindow(widget)


if __name__ == "__main__":
    app = widgets.app()
    widget = MdiArea()
    widget.set_background("green", "cross")
    le = widgets.LineEdit("test")
    le2 = widgets.LineEdit("test")
    widget.add(le)
    widget.add(le2)
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Callable
import dataclasses
import logging

from typing import Any

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Transformer:
    fn: Callable[[Any], Any]
    role: int | constants.ItemDataRole
    selector: Callable[[Any], bool]
    selector_role: int | constants.ItemDataRole


class SliceValueTransformationProxyModel(itemmodels.SliceIdentityProxyModel):
    """A proxy model which transforms cell contents based on a Callable.

    ### Example:
    ```py
    model = MyModel()
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier[::2, 2:].modify(xyz)
    table.show()
    ```

    or

    ```py
    indexer = (slice(None, None, 2), slice(2, None))
    proxy = itemmodels.SliceValueTransformationProxyModel(indexer=indexer)
    proxy.set_source_model(model)
    proxy.add_transformer(lambda x: x + "something", selector=lambda x: "abc" in x)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "value_transformation"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._transformers: list[Transformer] = []

    def clear(self):
        """Clear all transformers."""
        self._transformers = []

    def add_transformer(
        self,
        fn: Callable[[Any], Any],
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        selector: Callable[[Any], bool] | None = None,
        selector_role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        """Add a transformer for given role.

        If a selector callable is given, the transformer will only be applied if the
        selector returns True.
        The selector receives the content of given data role as an argument.

        Arguments:
            fn: Callable to transform data of given role
            role: Data role to transform
            selector: Callable to filter the indexes which should be transformed
            selector_role: Role to use for the selector callable
        """
        tr = Transformer(
            fn=fn,
            role=role,
            selector=selector,
            selector_role=selector_role,
        )
        self._transformers.append(tr)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        val = super().data(index, role)
        if not self.indexer_contains(index):
            return val
        for t in self._transformers:
            if t.role == role:
                selector_val = super().data(index, t.selector_role)
                if t.selector is None or t.selector(selector_val):
                    val = t.fn(selector_val)
        return val


if __name__ == "__main__":
    from prettyqt import debugging, widgets
    from prettyqt.qt import QtGui

    app = widgets.app()
    view = debugging.example_table()
    model = view.proxifier.get_proxy("value_transformation")
    model.add_transformer(lambda x: f"{x}test")
    model.add_transformer(
        lambda x: QtGui.QColor("red"),
        role=constants.BACKGROUND_ROLE,
        selector=lambda x: True,
    )
    model.set_column_slice(slice(0, 2))
    view.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


SelectionFlagStr = Literal[
    "none",
    "clear",
    "select",
    "deselect",
    "toggle",
    "current",
    "rows",
    "columns",
    "select_current",
    "toggle_current",
    "clear_and_select",
]

SELECTION_FLAG: bidict[SelectionFlagStr, core.QItemSelectionModel.SelectionFlag] = bidict(
    none=core.QItemSelectionModel.SelectionFlag.NoUpdate,
    clear=core.QItemSelectionModel.SelectionFlag.Clear,
    select=core.QItemSelectionModel.SelectionFlag.Select,
    deselect=core.QItemSelectionModel.SelectionFlag.Deselect,
    toggle=core.QItemSelectionModel.SelectionFlag.Toggle,
    current=core.QItemSelectionModel.SelectionFlag.Current,
    rows=core.QItemSelectionModel.SelectionFlag.Rows,
    columns=core.QItemSelectionModel.SelectionFlag.Columns,
    select_current=core.QItemSelectionModel.SelectionFlag.SelectCurrent,
    toggle_current=core.QItemSelectionModel.SelectionFlag.ToggleCurrent,
    clear_and_select=core.QItemSelectionModel.SelectionFlag.ClearAndSelect,
)


class ItemSelectionModel(core.ObjectMixin, core.QItemSelectionModel):
    """Keeps track of a view's selected items."""

    def __contains__(self, index: core.ModelIndex) -> bool:
        return self.isSelected(index)

    def set_current_index(self, index, flag: SelectionFlagStr):
        self.setCurrentIndex(index, SELECTION_FLAG[flag])


if __name__ == "__main__":
    model = ItemSelectionModel()



from __future__ import annotations

import re

from prettyqt import constants, core


class RegexMatchesModel(core.AbstractTableModel):
    HEADER = ["Start", "End", "Value", "Groups"]
    SUPPORTS = list[re.Match]

    def __init__(self, matches: list | None = None, **kwargs):
        super().__init__(**kwargs)
        self.matches = matches or []

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (re.Match(), *_):
                return True
            case _:
                return False

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def headerData(  # type: ignore
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self.HEADER[section]

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        item = self.matches[index.row()]
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return str(item.span()[0])
            case constants.DISPLAY_ROLE, 1:
                return str(item.span()[1])
            case constants.DISPLAY_ROLE, 2:
                return repr(item.group())
            case constants.DISPLAY_ROLE, 3:
                return str(len(item.groups()))
            case constants.USER_ROLE, _:
                return item.span()

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        """Override for AbstractitemModel base method."""
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        return 0 if parent.isValid() else len(self.matches)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TableView()
    teststring = "fjdjkslfj355fjdkj 55 55454"
    regex = r"(?:[0-9])*.([0-9])"
    teststring = "SH_6208069141055_BC000388_20110412101855"
    regex = r"(?:([a-z]{2,})_)?(\d+)_([a-z]{2,}\d+)_(\d+)$"
    compiled = re.compile(regex, re.IGNORECASE)
    matches = list(compiled.finditer(teststring))
    model = RegexMatchesModel(matches)

    view.set_model(model)
    dlg = widgets.Dialog()
    dlg.set_layout("horizontal")
    dlg.box.add_widget(view)
    dlg.show_blocking()
    view.resize(500, 300)
    app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.utils import get_repr


class RangeFilterProxyModel(core.SortFilterProxyModel):
    ID = "range_filter"

    def __init__(self, min_value=None, max_value=None, **kwargs):
        self._min_value = min_value
        self._max_value = max_value
        super().__init__(**kwargs)

    def __repr__(self):
        return get_repr(self, self.get_range)

    def filterAcceptsRow(self, source_row: int, parent: core.ModelIndex) -> bool:
        column = self.filterKeyColumn()
        role = self.filterRole()
        source_model = self.sourceModel()
        idx = source_model.index(source_row, column, parent)
        value = source_model.data(idx, role)
        if self._min_value is not None and value < self._min_value:
            return False
        return self._max_value is None or value <= self._max_value

    def set_min_value(self, value: float | int | None):
        self._min_value = value
        self.invalidateRowsFilter()

    def get_min_value(self) -> float | int:
        if self._min_value is None:
            return -float("inf")
        return self._min_value

    def set_max_value(self, value: float | int):
        self._max_value = value
        self.invalidateRowsFilter()

    def get_max_value(self) -> float | int:
        if self._max_value is None:
            return float("inf")
        return self._max_value

    def set_range(self, rng: tuple[float | int | None, float | int | None]):
        self._min_value, self._max_value = rng
        self.invalidateRowsFilter()

    def get_range(self) -> tuple[float | int | None, float | int | None]:
        return (self._min_value, self._max_value)

    min_value = core.Property(
        float,
        get_min_value,
        set_min_value,
        doc="Minimum allowed value",
    )
    max_value = core.Property(
        float,
        get_max_value,
        set_max_value,
        doc="Maximum allowed value",
    )


if __name__ == "__main__":
    from prettyqt import constants, widgets

    app = widgets.app()
    proxy = RangeFilterProxyModel(True, filter_role=constants.CHECKSTATE_ROLE)



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtLocation
from prettyqt.utils import bidict


ERROR = bidict(
    none=QtLocation.QPlaceReply.Error.NoError,
    place_does_not_exist=QtLocation.QPlaceReply.Error.PlaceDoesNotExistError,
    category_does_not_exist=QtLocation.QPlaceReply.Error.CategoryDoesNotExistError,
    communication=QtLocation.QPlaceReply.Error.CommunicationError,
    parse=QtLocation.QPlaceReply.Error.ParseError,
    permissions=QtLocation.QPlaceReply.Error.PermissionsError,
    unsupported=QtLocation.QPlaceReply.Error.UnsupportedError,
    bad_argument=QtLocation.QPlaceReply.Error.BadArgumentError,
    cancel=QtLocation.QPlaceReply.Error.CancelError,
    unknown=QtLocation.QPlaceReply.Error.UnknownError,
)

ErrorStr = Literal[
    "none",
    "place_does_not_exist",
    "category_does_not_exist",
    "communication",
    "parse",
    "permissions",
    "unsupported",
    "bad_argument",
    "cancel",
    "unknown",
]

TYPE = bidict(
    generic=QtLocation.QPlaceReply.Type.Reply,
    details=QtLocation.QPlaceReply.Type.DetailsReply,
    search=QtLocation.QPlaceReply.Type.SearchReply,
    search_suggestion=QtLocation.QPlaceReply.Type.SearchSuggestionReply,
    content=QtLocation.QPlaceReply.Type.ContentReply,
    id=QtLocation.QPlaceReply.Type.IdReply,
    match=QtLocation.QPlaceReply.Type.MatchReply,
)

TypeStr = Literal[
    "generic",
    "details",
    "search",
    "search_suggestion",
    "content",
    "id",
    "match",
]


class PlaceReplyMixin(core.ObjectMixin):
    def get_error(self) -> ErrorStr:
        """Return error type.

        Returns:
            Error type
        """
        return ERROR.inverse[self.error()]

    def get_type(self) -> TypeStr:
        """Return type.

        Returns:
            Type
        """
        return TYPE.inverse[self.type()]


class PlaceReply(PlaceReplyMixin, QtLocation.QPlaceReply):
    pass


if __name__ == "__main__":
    reply = PlaceReply()



from __future__ import annotations

import builtins

from prettyqt import core, widgets
from prettyqt.utils import helpers


class ListInput(widgets.LineEdit):
    value_changed = core.Signal(list)

    def __init__(
        self,
        value: list[float] | str = "",
        typ: type = int,
        allow_single: bool = False,
        object_name: str = "list_input",
        **kwargs,
    ):
        super().__init__(object_name=object_name, **kwargs)
        match typ:
            case builtins.int:
                self.set_validator("int_list", allow_single=allow_single)
            case builtins.float:
                self.set_validator("float_list", allow_single=allow_single)
            case _:
                raise ValueError(f"Invalid type {typ}")
        self.set_value(value)

    def get_value(self) -> list[float | int]:  # type: ignore[override]
        # TODO: is this best place to deal with exception?
        try:
            return helpers.string_to_num_array(self.text())
        except ValueError:
            return []

    def set_value(self, value: list[float] | str):
        if isinstance(value, list):
            value = str(value)[1:-1].replace(" ", "")
        self.set_text(value)


if __name__ == "__main__":
    app = widgets.app()
    widget = ListInput(typ=int)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import location
from prettyqt.utils import bidict


TYPE = bidict(
    none=location.QPlaceContent.Type.NoType,
    image=location.QPlaceContent.Type.ImageType,
    review=location.QPlaceContent.Type.ReviewType,
    editorial=location.QPlaceContent.Type.EditorialType,
    custom=location.QPlaceContent.Type.CustomType,
)

TypeStr = Literal["none", "image", "review", "editorial", "custom"]


class PlaceContentMixin:
    def get_type(self) -> TypeStr:
        """Return the visibility of the place.

        Returns:
            Place type
        """
        return TYPE.inverse[self.type()]

    def get_user(self) -> location.PlaceUser:
        return location.PlaceUser(self.user())

    def get_supplier(self) -> location.PlaceSupplier:
        return location.PlaceSupplier(self.supplier())


class PlaceContent(PlaceContentMixin, location.QPlaceContent):
    """Holds content about places."""


if __name__ == "__main__":
    content = PlaceContent()



from __future__ import annotations

from typing import Literal

from typing_extensions import Self

from prettyqt.qt import QtGui
from prettyqt.utils import bidict, get_repr, serializemixin


TRANSFORMATION_TYPE = bidict(
    none=QtGui.QTransform.TransformationType.TxNone,
    translate=QtGui.QTransform.TransformationType.TxTranslate,
    scale=QtGui.QTransform.TransformationType.TxScale,
    rotate=QtGui.QTransform.TransformationType.TxRotate,
    shear=QtGui.QTransform.TransformationType.TxShear,
    project=QtGui.QTransform.TransformationType.TxProject,
)

TransformationTypeStr = Literal[
    "none", "translate", "scale", "rotate", "shear", "project"
]


class Transform(serializemixin.SerializeMixin, QtGui.QTransform):
    def __repr__(self):
        return get_repr(
            self,
            self.m11(),
            self.m12(),
            self.m13(),
            self.m21(),
            self.m22(),
            self.m23(),
            self.m31(),
            self.m32(),
            self.m33(),
        )

    def __getitem__(self, value: tuple[int, int]) -> float:
        match value[0], value[1]:
            case 0, 0:
                return self.m11()
            case 0, 1:
                return self.m12()
            case 0, 2:
                return self.m13()
            case 1, 0:
                return self.m21()
            case 1, 1:
                return self.m22()
            case 1, 2:
                return self.m23()
            case 2, 0:
                return self.m31()
            case 2, 1:
                return self.m32()
            case 2, 2:
                return self.m33()
            case _:
                raise ValueError(f"Wrong value {value}")

    @property
    def _m11(self) -> float:
        return self.m11()

    @property
    def _m12(self) -> float:
        return self.m12()

    @property
    def _m13(self) -> float:
        return self.m13()

    @property
    def _m21(self) -> float:
        return self.m21()

    @property
    def _m22(self) -> float:
        return self.m22()

    @property
    def _m23(self) -> float:
        return self.m23()

    @property
    def _m31(self) -> float:
        return self.m31()

    @property
    def _m32(self) -> float:
        return self.m32()

    @property
    def _m33(self) -> float:
        return self.m33()

    __match_args__ = (
        "_m11",
        "_m12",
        "_m13",
        "_m21",
        "_m22",
        "_m23",
        "_m31",
        "_m32",
        "_m33",
    )

    @classmethod
    def clone_from(cls, transform: QtGui.QTransform) -> Self:
        return cls(
            transform.m11(),
            transform.m12(),
            transform.m13(),
            transform.m21(),
            transform.m22(),
            transform.m23(),
            transform.m31(),
            transform.m32(),
            transform.m33(),
        )

    def get_type(self) -> TransformationTypeStr:
        return TRANSFORMATION_TYPE.inverse[self.type()]


if __name__ == "__main__":
    transform = Transform()



from __future__ import annotations

from prettyqt.qt import QtQml
from prettyqt.utils import bidict


mod = QtQml.QQmlImageProviderBase

FLAGS = bidict(
    force_async_image_loading=mod.Flag.ForceAsynchronousImageLoading,
)


IMAGE_TYPE = bidict(
    image=mod.ImageType.Image,
    pixmap=mod.ImageType.Pixmap,
    texture=mod.ImageType.Texture,
    image_response=mod.ImageType.ImageResponse,
)


class QmlImageProviderBaseMixin:
    pass


class QmlImageProviderBase(QmlImageProviderBaseMixin, QtQml.QQmlImageProviderBase):
    """Used to register image providers in the QML engine."""



from __future__ import annotations

from prettyqt import core, gui, widgets
from prettyqt.utils import get_repr


class ElidedLabel(widgets.Frame):
    elision_changed = core.Signal(bool)

    def __init__(
        self,
        text: str = "",
        parent: widgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        self.elided = False
        self._text = text
        self.set_size_policy("expanding", "preferred")

    def __repr__(self):
        return get_repr(self, self.elided_text)

    def set_text(self, text: str):
        self._text = text
        self.update()

    def get_text(self) -> str:
        return self._text

    def paintEvent(self, event):
        super().paintEvent(event)
        with gui.Painter(self) as painter:
            metrics = painter.get_font_metrics()
            did_elide = False
            line_spacing = metrics.lineSpacing()
            y = 0
            layout = gui.TextLayout(self._text, painter.font())
            with layout.process_layout():
                while True:
                    line = layout.createLine()

                    if not line.isValid():
                        break

                    line.setLineWidth(self.width())
                    next_line_y = y + line_spacing

                    if self.height() >= next_line_y + line_spacing:
                        line.draw(painter, core.PointF(0, y))
                        y = next_line_y
                    else:
                        last_line = self._text[line.textStart() :]
                        elided_line = metrics.elided_text(
                            last_line, "right", self.width()
                        )
                        painter.drawText(0, y + metrics.ascent(), elided_line)
                        line = layout.createLine()
                        did_elide = line.isValid()
                        break
            if did_elide != self.elided:
                self.elided = did_elide
                self.elision_changed.emit(did_elide)

    elided_text = core.Property(str, get_text, set_text, doc="Unelided text")

    # def paintEvent(self, event):
    #     painter = gui.Painter(self)
    #     metrics = gui.FontMetrics(self.font())
    #     elided = metrics.elided_text(self.text(), "right", self.width())
    #     painter.drawText(self.rect(), self.alignment(), elided)


if __name__ == "__main__":
    app = widgets.app()
    widget = ElidedLabel("test")
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt.qt import QtDesigner


class DesignerCustomWidgetInterface(QtDesigner.QDesignerCustomWidgetInterface):
    """Enables Qt Designer to access and construct custom widgets."""


if __name__ == "__main__":
    interface = DesignerCustomWidgetInterface()



from __future__ import annotations

from prettyqt.qt import QtDesigner


class AbstractFormBuilderMixin:
    pass


class AbstractFormBuilder(AbstractFormBuilderMixin, QtDesigner.QAbstractFormBuilder):
    """Default implementation for classes that create user interfaces at run-time."""


if __name__ == "__main__":
    builder = AbstractFormBuilder()
    print(dir(builder))



from __future__ import annotations

import ast
import enum
import logging

from prettyqt import constants, core, gui, itemmodels
from prettyqt.utils import bidict


SOURCE_FONT = gui.Font.mono(as_qt=True)

NODE_MAP = bidict(
    {
        ast.Add: "+",
        ast.Sub: "-",
        ast.Mult: "*",
        ast.Div: "/",
        ast.Mod: "%",
        ast.Eq: "==",
        ast.In: "in",
        ast.Pow: "**",
        ast.MatMult: "@",
        ast.FloorDiv: "//",
        ast.BitAnd: "&",
        ast.BitOr: "|",
        ast.BitXor: "^",
        ast.Gt: ">",
        ast.GtE: ">=",
        ast.Lt: "<",
        ast.LtE: "<=",
        ast.Is: "is",
        ast.NotEq: "!=",
        ast.IsNot: "is not",
        ast.Not: "not ",
        ast.NotIn: "not in",
        ast.And: "and",
        ast.Or: "or",
        ast.Invert: "~",
    }
)


logger = logging.getLogger(__name__)


# AstTokens library would be worth out checking for this model.


class AstModel(itemmodels.TreeModel):
    """Tree model to display an Abstract syntax tree.

    The model shows a tree of all nodes from an
    [abstract syntax tree](https://docs.python.org/3/library/ast.html#node-classes)
    They are part of the builtin `ast` module.

    ### Example:
    ```py
    import ast
    view = TreeView()
    code = pathlib.Path(__file__).read_text()
    tree = ast.parse(code)
    model = AstModel(tree)
    view.set_model(model)
    ```
    """

    @core.Enum
    class Roles(enum.IntEnum):
        NodeRole = constants.USER_ROLE

    SUPPORTS = ast.AST
    HEADER = [
        "Node type",
        "Name",
        "Line range",
        "Column range",
        "Code segement",
        "Docstring",
    ]

    def __init__(self, ast_tree: ast.AST | str, **kwargs):
        super().__init__(None, **kwargs)
        self.ast_tree = None
        self.code = ""
        self.set_ast(ast_tree)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, ast.AST)

    def columnCount(self, parent=None):
        return len(self.HEADER) if self.ast_tree is not None else 0

    def set_ast(self, ast_tree: ast.AST | str = ""):
        """Set an AST tree for the model.

        Arguments:
            ast_tree: Abstract syntax tree
        """
        match ast_tree:
            case str():
                code = ast_tree
                try:
                    node = ast.parse(ast_tree)
                except SyntaxError as e:
                    logger.debug(f"caught {e!r} when building AST")
                    return
            case ast.AST():
                code = ast.unparse(ast_tree)
                node = ast.parse(code)  # makin a circle to make sure line numbers match.
            case _:
                raise TypeError(ast_tree)
        node = ast.fix_missing_locations(node)
        with self.reset_model():
            self.ast_tree = node
            self.set_root_item(node)
            self.code = code

    # def find_lineno(self, index):
    #     # not needed, thx to ast.fix_missing_locations
    #     while not hasattr(node := index.data(constants.USER_ROLE), "lineno"):
    #         index = index.parent()
    #     return node.lineno

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        match orientation, role, section:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.HEADER[section]
        return None

    def data(self, index: core.ModelIndex, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        node = self.data_by_index(index).obj
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return type(node).__name__
            case constants.DISPLAY_ROLE, 1:
                match node:
                    case _ if type(node) in NODE_MAP:
                        return NODE_MAP[type(node)]
                    case (
                        ast.Name(id=name)
                        | ast.arg(arg=name)
                        | ast.Constant(value=name)
                        | ast.alias(name=name)
                        | ast.ClassDef(name=name)
                        | ast.FunctionDef(name=name)
                    ):
                        return name
                    case str():
                        return node
            case constants.DISPLAY_ROLE, 2:
                if hasattr(node, "lineno"):
                    return (
                        f"{node.lineno} - {node.end_lineno}"
                        if node.lineno != node.end_lineno
                        else str(node.lineno)
                    )
            case constants.DISPLAY_ROLE, 3:
                if hasattr(node, "col_offset"):
                    return (
                        f"{node.col_offset} - {node.end_col_offset}"
                        if node.col_offset != node.end_col_offset
                        else str(node.col_offset)
                    )
            case constants.DISPLAY_ROLE, 4:
                return ast.get_source_segment(self.code, node)
            case constants.FONT_ROLE, 4 | 5:
                return SOURCE_FONT
            case constants.DISPLAY_ROLE, 5:
                try:
                    return ast.get_docstring(node)
                except TypeError:
                    return None
            case self.Roles.NodeRole, _:
                return node

    def _fetch_object_children(self, item: AstModel.TreeItem) -> list[AstModel.TreeItem]:
        return [self.TreeItem(obj=i) for i in ast.iter_child_nodes(item.obj)]

    def _has_children(self, item: AstModel.TreeItem) -> bool:
        if item.obj is None:
            return False
        return any(True for _ in ast.iter_child_nodes(item.obj))

    def rename_variable(
        self,
        old: str,
        new: str,
        root_tree: ast.AST | None = None,
        ignore: list[str] | None = None,
        scope: list[str] | None = None,
    ):
        if scope is None:
            scope = ["main"]
        if ignore is None:
            ignore = []
        if root_tree is None:
            root_tree = self.ast_tree
        for i in ast.iter_fields(root_tree):
            if not isinstance(a := getattr(root_tree, i), list):
                if a == old and not {*scope} & {*ignore}:
                    setattr(root_tree, i, new)
            n = a if isinstance(a, list) else [a]
            s = [root_tree.name] if type(root_tree).__name__.endswith("Def") else scope
            for j in n:
                if isinstance(j, ast.AST):
                    self.rename_variable(j, old, new, ignore, s)

    def get_variable_names(self):
        current_names = set()
        for node in ast.walk(self.ast_tree):
            match node:
                case ast.Name(id=name) | ast.arg(arg=name):
                    current_names.add(name)
        return current_names


if __name__ == "__main__":
    import pathlib

    from prettyqt import widgets

    app = widgets.app()
    with app.debug_mode():
        view = widgets.TreeView()
        code = pathlib.Path(__file__).read_text()
        tree = ast.parse(code)
        model = AstModel(tree)
        view.set_model(model)
        view.setEditTriggers(view.EditTrigger.AllEditTriggers)
        view.set_delegate("editor")
        view.resize(1000, 1000)
        view.show()
        with app.debug_mode():
            app.exec()



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.qt import QtCharts
from prettyqt.utils import get_repr


class CandlestickSet(core.ObjectMixin, QtCharts.QCandlestickSet):
    def __repr__(self):
        return get_repr(
            self, self.open(), self.high(), self.low(), self.close(), self.timestamp()
        )

    def get_pen(self) -> gui.Pen:
        return gui.Pen(self.pen())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())


if __name__ == "__main__":
    candlestickset = CandlestickSet(1, 3, 0, 2)



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWebEngineCore


class WebEngineHistoryItem(QtWebEngineCore.QWebEngineHistoryItem):
    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def get_icon_url(self) -> core.Url:
        return core.Url(self.iconUrl())

    def get_last_visited(self) -> core.DateTime:
        return core.DateTime(self.lastVisited())



from __future__ import annotations

from typing import Literal

from prettyqt import network
from prettyqt.qt import QtCore
from prettyqt.utils import bidict


CapabilityStr = Literal[
    "tunneling",
    "listening",
    "udp_tunneling",
    "caching",
    "host_name_lookup",
    "sctp_tunneling",
    "sctp_listening",
]

CAPABILITIES: bidict[CapabilityStr, network.QNetworkProxy.Capability] = bidict(
    tunneling=network.QNetworkProxy.Capability.TunnelingCapability,
    listening=network.QNetworkProxy.Capability.ListeningCapability,
    udp_tunneling=network.QNetworkProxy.Capability.UdpTunnelingCapability,
    caching=network.QNetworkProxy.Capability.CachingCapability,
    host_name_lookup=network.QNetworkProxy.Capability.HostNameLookupCapability,
    sctp_tunneling=network.QNetworkProxy.Capability.SctpTunnelingCapability,
    sctp_listening=network.QNetworkProxy.Capability.SctpListeningCapability,
)

ProxyTypeStr = Literal[
    "none",
    "default",
    "socks5",
    "http",
    "http_caching",
    "ftp_caching",
]

PROXY_TYPES: bidict[ProxyTypeStr, network.QNetworkProxy.ProxyType] = bidict(
    none=network.QNetworkProxy.ProxyType.NoProxy,
    default=network.QNetworkProxy.ProxyType.DefaultProxy,
    socks5=network.QNetworkProxy.ProxyType.Socks5Proxy,
    http=network.QNetworkProxy.ProxyType.HttpProxy,
    http_caching=network.QNetworkProxy.ProxyType.HttpCachingProxy,
    ftp_caching=network.QNetworkProxy.ProxyType.FtpCachingProxy,
)


class NetworkProxy(network.QNetworkProxy):
    """Network layer proxy."""

    def get_capabilities(self) -> list[CapabilityStr]:
        return CAPABILITIES.get_list(self.capabilities())

    def set_capabilities(self, *capability: CapabilityStr):
        flags = CAPABILITIES.merge_flags(capability)
        self.setCapabilities(flags)

    def get_header(
        self,
        name: network.networkrequest.KnownHeaderStr | network.NetworkRequest.KnownHeaders,
    ) -> str:
        return self.header(network.networkrequest.KNOWN_HEADER.get_enum_value(name))

    def set_header(
        self,
        name: network.networkrequest.KnownHeaderStr | network.NetworkRequest.KnownHeaders,
        value: str,
    ):
        self.setHeader(network.networkrequest.KNOWN_HEADER.get_enum_value(name), value)

    def get_headers(self) -> dict[str, str]:
        return {
            h.data().decode(): self.rawHeader(h).data().decode()
            for h in self.rawHeaderList()
        }

    def set_headers(self, headers: dict[str, str]):
        for k, v in headers.items():
            self.setRawHeader(
                QtCore.QByteArray(k.encode()), QtCore.QByteArray(v.encode())
            )

    def set_type(self, typ: ProxyTypeStr | network.QNetworkProxy.ProxyType):
        """Set proxy type.

        Args:
            typ: proxy type
        """
        self.setType(PROXY_TYPES.get_enum_value(typ))

    def get_type(self) -> ProxyTypeStr:
        """Get the proxy type.

        Returns:
            type
        """
        return PROXY_TYPES.inverse[self.type()]



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtHelp


class HelpFilterSettingsWidget(widgets.WidgetMixin, QtHelp.QHelpFilterSettingsWidget):
    """Widget that allows for creating, editing and removing filters."""


if __name__ == "__main__":
    app = widgets.app()
    widget = HelpFilterSettingsWidget()
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Callable, Container

from prettyqt import core
from prettyqt.utils import get_repr, helpers


class SubsetFilterProxyModel(core.SortFilterProxyModel):
    """A FilterProxyModel to filter based on slices, ranges, indexes or Callables.

    ### Example

    ```py
    proxy = itemmodels.SubsetFilterProxyModel()
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```

    !!! note
        If you only need filtering based on slices or a single column / row,
        the SliceFilterProxymodel should be preferred for performance reasons.
    """

    ID = "subset"

    def __init__(
        self,
        row_filter: slice | range | int | Container[int] | Callable | None,
        column_filter: slice | range | int | Container[int] | Callable | None,
        **kwargs,
    ):
        self.row_filter = row_filter
        self.column_filter = column_filter
        super().__init__(**kwargs)

    def __repr__(self):
        return get_repr(self, self.row_filter, self.column_filter)

    def filterAcceptsColumn(self, source_column: int, parent: core.ModelIndex) -> bool:
        match self.column_filter:
            case slice() | range():
                return helpers.is_in_slice(self.column_filter, source_column)
            case int():
                return source_column == self.column_filter
            case Container():
                return source_column in self.column_filter
            case Callable():
                return self.column_filter(source_column)
            case None:
                return True
            case _:
                raise ValueError(self.column_filter)

    def filterAcceptsRow(self, source_row: int, parent: core.ModelIndex) -> bool:
        match self.row_filter:
            case slice() | range():
                return helpers.is_in_slice(self.row_filter, source_row)
            case int():
                return source_row == self.row_filter
            case Container():
                return source_row in self.row_filter
            case Callable():
                return self.row_filter(source_row)
            case None:
                return True
            case _:
                raise ValueError(self.row_filter)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    proxy = SubsetFilterProxyModel(1, 2)
